"""API -> TC"""
# standard library
from typing import TYPE_CHECKING, Dict, List, Union

# first-party
from tcex.api.tc.ti_transform import TiTransform, TiTransforms
from tcex.api.tc.ti_transform.model import GroupTransformModel, IndicatorTransformModel
from tcex.api.tc.v2.v2 import V2
from tcex.api.tc.v3.v3 import V3
from tcex.backports import cached_property

if TYPE_CHECKING:
    # third-party
    from requests import Session

    # first-party
    from tcex.input.input import Input


class TC:
    """API -> TC

    Args:
        inputs: An instance of the Input class.
        session_tc: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, inputs: 'Input', session_tc: 'Session'):
        """Initialize Class properties."""
        self.inputs = inputs
        self.session_tc = session_tc

    @staticmethod
    def group_transform(transform: Dict) -> 'GroupTransformModel':
        """Return a group transform model."""
        return GroupTransformModel(**transform)

    @staticmethod
    def indicator_transform(transform: Dict) -> 'IndicatorTransformModel':
        """Return a indicator transform model."""
        return IndicatorTransformModel(**transform)

    @staticmethod
    def ti_transform(
        ti_dict: dict,
        transforms: List[Union['GroupTransformModel', 'IndicatorTransformModel']],
    ) -> 'TiTransform':
        """Return an instance of TI Transform class."""
        return TiTransform(ti_dict, transforms)

    @staticmethod
    def ti_transforms(
        ti_dict: List[dict],
        transforms: List[Union['GroupTransformModel', 'IndicatorTransformModel']],
    ) -> 'TiTransform':
        """Return an instance of TI Transforms class."""
        return TiTransforms(ti_dict, transforms)

    @cached_property
    def v2(self) -> 'V2':
        """Return a case management instance."""
        return V2(self.inputs, self.session_tc)

    @cached_property
    def v3(self) -> 'V3':
        """Return a case management instance."""
        return V3(self.session_tc)



"""ThreatConnect TQL"""
# standard library
from typing import List, Optional

# first-party
from tcex.api.tc.v3.tql.tql_type import TqlType


class Tql:
    """ThreatConnect TQL"""

    def __init__(self):
        """Initialize Class Properties"""
        self._filters = []
        self.raw_tql = None

    @property
    def as_str(self):
        """Convert the TQL obj to a string"""
        filters = []
        for tql_filter in self.filters:
            # keywords are all one work (e.g. task_id should be taskid)
            keyword = tql_filter.get('keyword').replace('_', '')
            value = tql_filter.get('value')
            try:
                filters.append(f'''{keyword}({value._tql.as_str})''')
            except Exception:
                if isinstance(value, List):
                    if tql_filter.get('type') == TqlType.INTEGER:
                        value = [str(int_) for int_ in value]
                    elif tql_filter.get('type') == TqlType.STRING:
                        value = [f'"{str(str_)}"' for str_ in value]
                    value = ','.join(value)
                    value = f'({value})'
                elif tql_filter.get('type') == TqlType.STRING:
                    value = f'"{value}"'
                filters.append(f'''{keyword} {tql_filter.get('operator').value} {value}''')

        return ' and '.join(filters)

    @property
    def filters(self) -> List[dict]:
        """Return the filters"""
        return self._filters

    @filters.setter
    def filters(self, filters: List[dict]):
        """Set the filters"""
        self._filters = filters

    def add_filter(
        self, keyword: str, operator: str, value: str, type_: Optional[TqlType] = TqlType.STRING
    ):
        """Add a filter to the current obj

        Args:
            keyword: the field to search on
            operator: the operator to use
            value: the value to compare
            type_: How to treat the value (defaults to String)
        """
        self.filters.append(
            {'keyword': keyword, 'operator': operator, 'value': value, 'type': type_}
        )

    def set_raw_tql(self, tql: str):
        """Set a raw TQL filter"""
        self.raw_tql = tql



#!/usr/bin/env python
"""TcEx Framework Validate Module."""
# standard library
import ast
import importlib
import json
import os
import sys
import traceback
from collections import deque
from pathlib import Path
from typing import Dict, Union

# third-party
import colorama as c
from pydantic import ValidationError
from stdlib_list import stdlib_list

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.app_config.job_json import JobJson
from tcex.app_config.layout_json import LayoutJson
from tcex.app_config.permutation import Permutation
from tcex.app_config.tcex_json import TcexJson
from tcex.bin.bin_abc import BinABC

try:
    # standard library
    import sqlite3
except ModuleNotFoundError:
    # this module is only required for certain CLI commands
    pass


class Validate(BinABC):
    """Validate syntax, imports, and schemas.

    * Python and JSON file syntax
    * Python import modules
    * install.json schema
    * layout.json schema
    """

    def __init__(self, ignore_validation: bool):
        """Initialize Class properties."""
        super().__init__()
        self.permutations = Permutation()
        self.ignore_validation = ignore_validation

        # class properties
        self._app_packages = []
        self._install_json_schema = None
        self._layout_json_schema = None
        self.config = {}
        self.ij = InstallJson()
        self.invalid_json_files = []
        self.lj = LayoutJson()
        self.tj = TcexJson()

        # initialize validation data
        self.validation_data = self._validation_data

    @property
    def _validation_data(self) -> Dict[str, list]:
        """Return structure for validation data."""
        return {
            'errors': [],
            'fileSyntax': [],
            'layouts': [],
            'moduleImports': [],
            'schema': [],
            'feeds': [],
        }

    def _check_node_import(self, node: Union[ast.Import, ast.ImportFrom], filename: str):
        """."""
        if isinstance(node, ast.Import):
            for n in node.names:
                m = n.name.split('.')[0]
                if not self.check_import_stdlib(m):
                    m_status = self.check_imported(m)
                    if not m_status:
                        self.validation_data['errors'].append(
                            f'Module validation failed for {filename} '
                            f'module "{m}" could not be imported).'
                        )
                    self.validation_data['moduleImports'].append(
                        {'filename': filename, 'module': m, 'status': m_status}
                    )
        elif isinstance(node, ast.ImportFrom):
            m = node.module.split('.')[0]
            if not self.check_import_stdlib(m):
                m_status = self.check_imported(m)
                if not m_status:
                    self.validation_data['errors'].append(
                        f'Module validation failed for {filename} '
                        f'module "{m}" could not be imported).'
                    )
                self.validation_data['moduleImports'].append(
                    {'filename': filename, 'module': m, 'status': m_status}
                )

    def check_imports(self):
        """Check the projects top level directory for missing imports.

        This method will check only files ending in **.py** and does not handle imports validation
        for sub-directories.
        """
        for filename in sorted(os.listdir(self.app_path)):
            if not filename.endswith('.py'):
                continue

            fq_path = os.path.join(self.app_path, filename)
            with open(fq_path, 'rb') as f:
                # TODO: [low] is there a better way?
                code_lines = deque([(f.read(), 1)])

                while code_lines:
                    code, _ = code_lines.popleft()  # pylint: disable=unused-variable
                    try:
                        parsed_code = ast.parse(code)
                        for node in ast.walk(parsed_code):
                            self._check_node_import(node, filename)
                    except SyntaxError:
                        pass

    @staticmethod
    def check_import_stdlib(module: str) -> bool:
        """Check if module is in Python stdlib.

        Args:
            module: The name of the module to check.

        Returns:
            bool: Returns True if the module is in the stdlib or template.
        """
        if (
            module in stdlib_list('3.6')
            or module in stdlib_list('3.7')
            or module in stdlib_list('3.8')
            or module
            in ['app', 'args', 'base_app_input', 'job_app', 'playbook_app', 'run', 'service_app']
        ):
            return True
        return False

    @staticmethod
    def check_imported(module: str) -> bool:
        """Check whether the provide module can be imported (package installed).

        Args:
            module: The name of the module to check availability.

        Returns:
            bool: True if the module can be imported, False otherwise.
        """
        try:
            del sys.modules[module]
        except (AttributeError, KeyError):
            pass

        # https://docs.python.org/3/library/importlib.html#checking-if-a-module-can-be-imported
        find_spec = importlib.util.find_spec(module)
        found = find_spec is not None
        if found is True:
            # if dist-packages|site-packages in module_path the import doesn't count
            try:
                if 'dist-packages' in find_spec.origin:
                    found = False
            except TypeError:
                pass

            try:
                if 'site-packages' in find_spec.origin:
                    found = False
            except TypeError:
                pass
        return found

    def check_install_json(self):
        """Check all install.json files for valid schema."""
        if 'install.json' in self.invalid_json_files:
            return

        status = True
        try:
            self.ij.model
        except ValidationError as ex:
            self.invalid_json_files.append(self.ij.fqfn.name)
            status = False
            for error in json.loads(ex.json()):
                location = [str(location) for location in error.get('loc')]
                self.validation_data['errors'].append(
                    '''Schema validation failed for install.json. '''
                    f'''{error.get('msg')}: {' -> '.join(location)}'''
                )
        except ValueError:
            # any JSON decode error will be caught during syntax validation
            return

        self.validation_data['schema'].append({'filename': self.ij.fqfn.name, 'status': status})

    def check_job_json(self):
        """Validate feed files for feed job apps."""
        if 'install.json' in self.invalid_json_files:
            # can't proceed if install.json can't be read
            return

        if not self.ij.model.feeds:
            # if no jobs, skip!
            return

        # use developer defined app version (deprecated) or package_version from InstallJson model
        app_version = self.tj.model.package.app_version or self.ij.model.package_version
        program_name = (f'''{self.tj.model.package.app_name}_{app_version}''').replace('_', ' ')
        status = True
        for feed in self.ij.model.feeds or []:
            if feed.job_file in self.invalid_json_files:
                # no need to check if schema if json is invalid
                continue

            jj = JobJson(filename=feed.job_file)

            # validate the job file exists
            if not jj.fqfn.is_file():
                self.validation_data['errors'].append(
                    f'''Schema validation failed for {feed.job_file}. '''
                    f'''The job.json file could not be found.'''
                )
                continue

            try:
                # validate the schema
                jj.model
            except ValidationError as ex:
                status = False
                for error in json.loads(ex.json()):
                    location = [str(location) for location in error.get('loc')]
                    self.validation_data['errors'].append(
                        f'''Schema validation failed for {feed.job_file}. '''
                        f'''{error.get('msg')}: {' -> '.join(location)}'''
                    )

            # validate program name
            if status is True and jj.model.program_name != program_name:
                status = False
                self.validation_data['errors'].append(
                    f'''Schema validation failed for {feed.job_file}. '''
                    f'''The job.json programName {jj.model.program_name} != {program_name}.'''
                )

            # validate program version
            if status is True and jj.model.program_version != self.ij.model.program_version:
                status = False
                self.validation_data['errors'].append(
                    f'''Schema validation failed for {feed.job_file}. The job.json program'''
                    f'''Version {jj.model.program_version} != {self.ij.model.program_version}.'''
                )

            self.validation_data['schema'].append({'filename': feed.job_file, 'status': status})

    def check_layout_json(self):
        """Check all layout.json files for valid schema."""
        if not self.lj.has_layout or 'layout.json' in self.invalid_json_files:
            return

        status = True
        try:
            self.lj.model
        except ValidationError as ex:
            self.invalid_json_files.append(self.ij.fqfn.name)
            status = False
            for error in json.loads(ex.json()):
                location = [str(location) for location in error.get('loc')]
                self.validation_data['errors'].append(
                    f'''Schema validation failed for layout.json. '''
                    f'''{error.get('msg')}: {' -> '.join(location)}'''
                )
        except ValueError:
            # any JSON decode error will be caught during syntax validation
            return

        self.validation_data['schema'].append({'filename': self.lj.fqfn.name, 'status': status})

        if status is True:
            self.check_layout_params()

    def check_layout_params(self):
        """Check that the layout.json is consistent with install.json.

        The layout.json files references the params.name from the install.json file.  The method
        will validate that no reference appear for inputs in install.json that don't exist.
        """
        # do not track hidden or serviceConfig inputs as they should not be in layouts.json
        ij_input_names = list(self.ij.model.filter_params(service_config=False, hidden=False))
        # pylint: disable=no-member
        ij_output_names = [o.name for o in self.ij.model.playbook.output_variables]

        # Check for duplicate inputs
        for name in self.ij.validate.validate_duplicate_input():
            self.validation_data['errors'].append(
                f'Duplicate input name found in install.json ({name})'
            )
            status = False

        # Check for duplicate sequence numbers
        for sequence in self.ij.validate.validate_duplicate_sequence():
            self.validation_data['errors'].append(
                f'Duplicate sequence number found in install.json ({sequence})'
            )
            status = False

        # Check for duplicate outputs variables
        for output in self.ij.validate.validate_duplicate_output():
            self.validation_data['errors'].append(
                f'Duplicate output variable name found in install.json ({output})'
            )
            status = False

        if 'sqlite3' in sys.modules:
            # create temporary inputs tables
            self.permutations.db_create_table(self.permutations._input_table, ij_input_names)

        # inputs
        status = True
        for i in self.lj.model.inputs:
            for p in i.parameters:
                if p.name not in ij_input_names:
                    # update validation data errors
                    self.validation_data['errors'].append(
                        'Layouts input.parameters[].name validations failed '
                        f'''("{p.name}" is defined in layout.json, '''
                        'but hidden or not found in install.json).'
                    )
                    status = False
                else:
                    # any item in list afterwards is a problem
                    ij_input_names.remove(p.name)

                if 'sqlite3' in sys.modules:
                    if p.display:
                        display_query = (
                            f'''SELECT * FROM {self.permutations._input_table}'''  # nosec
                            f''' WHERE {p.display}'''
                        )
                        try:
                            self.permutations.db_conn.execute(display_query.replace('"', ''))
                        except sqlite3.Error:
                            self.validation_data['errors'].append(
                                '''Layouts input.parameters[].display validations failed '''
                                f'''("{p.display}" query is an invalid statement).'''
                            )
                            status = False

        # update validation data for module
        self.validation_data['layouts'].append({'params': 'inputs', 'status': status})

        if ij_input_names:
            input_names = ','.join(ij_input_names)
            # update validation data errors
            self.validation_data['errors'].append(
                f'Layouts input.parameters[].name validations failed ("{input_names}" '
                'values from install.json were not included in layout.json.'
            )
            status = False

        # outputs
        status = True
        for o in self.lj.model.outputs:
            if o.name not in ij_output_names:
                # update validation data errors
                self.validation_data['errors'].append(
                    f'''Layouts output validations failed ({o.name} is defined '''
                    '''in layout.json, but not found in install.json).'''
                )
                status = False

            if 'sqlite3' in sys.modules:
                if o.display:
                    display_query = (
                        f'''SELECT * FROM {self.permutations._input_table} '''  # nosec
                        f'''WHERE {o.display}'''
                    )
                    try:
                        self.permutations.db_conn.execute(display_query.replace('"', ''))
                    except sqlite3.Error:
                        self.validation_data['errors'].append(
                            f"""Layouts outputs.display validations failed ("{o.display}" """
                            f"""query is an invalid statement)."""
                        )
                        status = False

        # update validation data for module
        self.validation_data['layouts'].append({'params': 'outputs', 'status': status})

    def check_syntax(self, app_path=None):
        """Run syntax on each ".py" and ".json" file.

        Args:
            app_path (str, optional): The path of Python files.
        """
        fqpn = Path(app_path or os.getcwd())

        for fqfn in sorted(fqpn.iterdir()):
            error = None
            status = True
            if fqfn.name.endswith('.py'):
                try:
                    with fqfn.open(mode='rb') as fh:
                        ast.parse(fh.read(), filename=fqfn.name)
                except SyntaxError:
                    status = False

                    # cleanup output
                    e = []
                    for line in traceback.format_exc().split('\n')[-5:-2]:
                        e.append(line.strip())
                    error = ' '.join(e)

            elif fqfn.name.endswith('.json'):
                try:
                    with fqfn.open() as fh:
                        json.load(fh)
                except ValueError as e:
                    # update tracker for common files
                    self.invalid_json_files.append(fqfn.name)
                    status = False
                    error = e
            else:
                # skip unsupported file types
                continue

            if error:
                # update validation data errors
                self.validation_data['errors'].append(
                    f'Syntax validation failed for {fqfn.name} ({error}).'
                )

            # store status for this file
            self.validation_data['fileSyntax'].append({'filename': fqfn.name, 'status': status})

    def interactive(self):
        """[App Builder] Run in interactive mode."""
        while True:
            line = sys.stdin.readline().strip()
            if line == 'quit':
                sys.exit()
            elif line == 'validate':
                self.check_syntax()
                self.check_imports()
                self.check_install_json()
                self.check_layout_json()
                self.check_job_json()
                self.print_json()

            # reset validation_data
            self.validation_data = self._validation_data

    def print_json(self):
        """[App Builder] Print JSON output."""
        print(json.dumps({'validation_data': self.validation_data}))

    # TODO: [low] switch to typer echo?
    def _print_file_syntax_results(self):
        """Print file syntax results."""
        if self.validation_data.get('fileSyntax'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated File Syntax:')
            print(f'''{c.Style.BRIGHT}{'File:'!s:<60}{'Status:'!s:<25}''')
            for f in self.validation_data.get('fileSyntax'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f"{f.get('filename')!s:<60}{status_color}{status_value!s:<25}")

    def _print_imports_results(self):
        """Print import results."""
        if self.validation_data.get('moduleImports'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Imports:')
            print(f'''{c.Style.BRIGHT}{'File:'!s:<30}{'Module:'!s:<30}{'Status:'!s:<25}''')
            for f in self.validation_data.get('moduleImports'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(
                    f'''{f.get('filename')!s:<30}{c.Fore.WHITE}'''
                    f'''{f.get('module')!s:<30}{status_color}{status_value!s:<25}'''
                )

    def _print_schema_results(self):
        """Print schema results."""
        if self.validation_data.get('schema'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Schema:')
            print(f'''{c.Style.BRIGHT}{'File:'!s:<60}{'Status:'!s:<25}''')
            for f in self.validation_data.get('schema'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f'''{f.get('filename')!s:<60}{status_color}{status_value!s:<25}''')

    def _print_layouts_results(self):
        """Print layout results."""
        if self.validation_data.get('layouts'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Layouts:')
            print(f'''{c.Style.BRIGHT}{'Params:'!s:<60}{'Status:'!s:<25}''')
            for f in self.validation_data.get('layouts'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f"{f.get('params')!s:<60}{status_color}{status_value!s:<25}")

    def _print_feed_results(self):
        """Print feed results."""
        if self.validation_data.get('feeds'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Feed Jobs:')
            print(f'''{c.Style.BRIGHT}{'Feeds:'!s:<60}{'Status:'!s:<25}''')
            for f in self.validation_data.get('feeds'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f"{f.get('name')!s:<60}{status_color}{status_value!s:<25}")

    def _print_errors(self):
        """Print errors results."""
        if self.validation_data.get('errors'):
            print('\n')  # separate errors from normal output
        for error in self.validation_data.get('errors'):
            # print all errors
            print(f'* {c.Fore.RED}{error}')

            # ignore exit code
            if not self.ignore_validation:
                self.exit_code = 1

    def print_results(self):
        """Print results."""
        # Validating Syntax
        self._print_file_syntax_results()

        # Validating Imports
        self._print_imports_results()

        # Validating Schema
        self._print_schema_results()

        # Validating Layouts
        self._print_layouts_results()

        # Validating Feed Job Definition Files
        self._print_feed_results()

        self._print_errors()

    @staticmethod
    def status_color(status) -> str:
        """Return the appropriate status color."""
        return c.Fore.GREEN if status else c.Fore.RED

    @staticmethod
    def status_value(status) -> str:
        """Return the appropriate status color."""
        return 'passed' if status else 'failed'



"""Install JSON Model"""
# pylint: disable=no-self-argument,no-self-use; noqa: N805
# standard library
from typing import List, Optional

# third-party
from pydantic import BaseModel, validator
from semantic_version import Version


class TemplateConfigModel(BaseModel):
    """App Template Config Model."""

    contributor: str
    description: str
    name: str
    summary: str
    template_files: List[str]
    template_parents: Optional[List[str]] = []
    type: str
    version: Version

    @validator('version', pre=True)
    def version_validator(cls, v):
        """Return a version object for "version" fields."""
        if v is not None:
            return Version(v)
        return v

    class Config:
        """DataModel Config"""

        arbitrary_types_allowed = True
        json_encoders = {Version: lambda v: str(v)}  # pylint: disable=unnecessary-lambda
        validate_assignment = True



"""Indicator Entity Field (Model) Type"""
# standard library
from typing import TYPE_CHECKING, List

# third-party
from pydantic import validator

# first-party
from tcex.api.tc.utils.threat_intel_utils import ThreatIntelUtils
from tcex.input.field_types.exception import InvalidEmptyValue, InvalidEntityType
from tcex.input.field_types.tc_entity import TCEntity
from tcex.pleb.registry import registry

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField


# pylint: disable=no-self-argument, no-self-use
class IndicatorEntity(TCEntity):
    """Indicator Entity Field (Model) Type"""

    @validator('type')
    def is_empty(cls, value: str, field: 'ModelField') -> str:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    @validator('type')
    def is_type(cls, value: str, field: 'ModelField') -> str:
        """Validate that the entity is of Indicator type."""
        ti_utils = ThreatIntelUtils(session_tc=registry.session_tc)
        if value not in ti_utils.indicator_types:
            raise InvalidEntityType(field_name=field.name, entity_type='Indicator', value=value)
        return value


def indicator_entity(indicator_types: List[str] = None) -> type:
    """Return custom model for Indicator Entity."""

    class CustomIndicatorEntity(IndicatorEntity):
        """Indicator Entity Field (Model) Type"""

        @validator('type', allow_reuse=True)
        def is_empty(cls, value: str, field: 'ModelField') -> str:
            """Validate that the value is a non-empty string."""
            if isinstance(value, str) and value.replace(' ', '') == '':
                raise InvalidEmptyValue(field_name=field.name)
            return value

        @validator('type', allow_reuse=True)
        def is_type(cls, value: str, field: 'ModelField') -> str:
            """Validate that the entity is of a specific Indicator type."""
            if value.lower() not in [i.lower() for i in indicator_types]:
                raise InvalidEntityType(
                    field_name=field.name, entity_type=str(indicator_types), value=value
                )
            return value

    return CustomIndicatorEntity


AddressEntity: IndicatorEntity = indicator_entity(indicator_types=['Address'])
EmailAddressEntity: IndicatorEntity = indicator_entity(indicator_types=['EmailAddress'])
HostEntity: IndicatorEntity = indicator_entity(indicator_types=['Host'])
FileEntity: IndicatorEntity = indicator_entity(indicator_types=['File'])
UrlEntity: IndicatorEntity = indicator_entity(indicator_types=['URL'])



#!/usr/bin/env python
"""TcEx Dependencies Command"""
# standard library
import hashlib
import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

# third-party
import yaml
from pydantic import ValidationError
from requests import Session
from requests.auth import HTTPBasicAuth
from tinydb import Query, TinyDB

# first-party
from tcex.app_config.models import TemplateConfigModel
from tcex.backports import cached_property
from tcex.bin.bin_abc import BinABC

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from requests import Response


class Template(BinABC):
    """Install dependencies for App."""

    def __init__(self):
        """Initialize class properties."""
        super().__init__()

        # properties
        self.base_url = os.getenv(
            'TCEX_TEMPLATE_URL',
            'https://api.github.com/repos/ThreatConnect-Inc/tcex-app-templates',
        )
        self.base_raw_url = 'https://raw.githubusercontent.com/ThreatConnect-Inc/tcex-app-templates'
        self.errors = False
        self.password = os.getenv('GITHUB_PAT')
        self.template_configs = {}
        self.template_data = {}
        self.template_manifest_fqfn = Path('.template_manifest.json')
        self.username = os.getenv('GITHUB_USER')

    @cached_property
    def cache_valid(self) -> bool:
        """Return the current commit sha for the tcex-app-templates project."""
        if self.project_sha is None:
            return False

        # get the store sha
        stored_sha = self.db_get_sha(self.project_sha)

        # if sha values don't match upsert and return False
        if stored_sha != self.project_sha:
            # upsert current sha
            self.db_add_sha(self.project_sha)
            return False

        # cache is valid
        return True

    def contents(
        self,
        branch: str,
        type_: str,
        template: Optional[str] = None,
        app_builder: Optional[bool] = False,
    ) -> dict:
        """Yield template contents."""
        url = f'{self.base_url}/contents/{type_}/{template}'
        if template is None:
            url = f'{self.base_url}/contents/{type_}'
        elif template == '_app_common':
            url = f'{self.base_url}/contents/{template}'

        params = {'ref': branch}
        r: 'Response' = self.session.get(url, params=params)
        if not r.ok:
            self.log.error(
                f'action=get-contents, url={r.request.url}, '
                f'status_code={r.status_code}, headers={r.headers}, '
                f'response={r.text or r.reason}'
            )
            self.errors = True
        else:
            for content in r.json():
                # exclusion - this file is only needed for building App Builder templates
                if content.get('name') == '.appbuilderconfig' and app_builder is False:
                    continue

                # exclusions - files that should not be part of the App
                if content.get('name') in ['.gitignore', 'template.yaml']:
                    continue

                # rename gitignore to .gitignore
                if content.get('name') == 'gitignore':
                    content['name'] = '.gitignore'

                yield content

    @cached_property
    def db(self) -> 'TinyDB':
        """Return db instance."""
        try:
            return TinyDB(os.path.join(self.cli_out_path, 'tcex.json'))
        except Exception:
            return None

    def db_add_config(self, config: TemplateConfigModel):
        """Add a config to the DB."""
        Config = Query()
        try:
            if config.name == '_app_common':
                config.type = '_app_common'
            self.db.upsert(
                json.loads(config.json()),
                (Config.type == config.type) & (Config.name == config.name),
            )
        except Exception as ex:
            self.log.error(f'Failed inserting config in db ({ex}).')
            self.errors = True

    def db_add_sha(self, sha: str):
        """Add a config to the DB."""
        SHA = Query()
        try:
            self.db.upsert({'sha': sha}, SHA.sha.exists())
        except Exception as ex:
            self.log.error(f'Failed inserting config in db ({ex}).')
            self.errors = True

    def db_get_config(self, type_: str, template: str) -> Optional['TemplateConfigModel']:
        """Get a config from the DB."""
        Config = Query()
        try:
            if template == '_app_common':
                type_ = '_app_common'
            config = self.db.search((Config.type == type_) & (Config.name == template))

            if config:
                return TemplateConfigModel(**config[0])
            return None
        except Exception as ex:
            self.log.error(f'Failed retrieving config from db ({ex}).')
            self.errors = True
            return None

    def db_get_sha(self, sha: str) -> Optional[str]:
        """Get repo SHA from the DB."""
        SHA = Query()
        try:
            sha = self.db.search(SHA.sha == sha)
            if sha:
                return sha[0].get('sha')
            return None
        except Exception as ex:
            self.log.error(f'Failed inserting config in db ({ex}).')
            self.errors = True
            return None

    def download_template_file(self, item: dict):
        """Download the provided source file to the provided destination."""
        download_url = item.get('download_url')
        r: 'Response' = self.session.get(
            download_url, allow_redirects=True, headers={'Cache-Control': 'no-cache'}
        )
        if not r.ok:
            self.log.error(
                f'action=download-template-file, url={r.request.url}, '
                f'status_code={r.status_code}, headers={r.headers}, '
                f'response={r.text or r.reason}'
            )

        # get the relative path to the file and create the parent directory if it does not exist
        destination = item.get('relative_path')
        if destination.parent.exists() is False:
            destination.parent.mkdir(parents=True, exist_ok=True)
        destination.open(mode='wb').write(r.content)
        self.log.info(f'action=download-template-file, file={destination}')

        # update manifest, using the path as the key for uniqueness
        self.template_manifest[item.get('path')]['md5'] = self.file_hash(destination)

    @staticmethod
    def file_hash(fqfn: Path) -> str:
        """Return the file hash."""
        with fqfn.open(mode='rb') as fh:
            md5 = hashlib.md5()  # nosec
            while True:
                chunk = fh.read(8192)
                if not chunk:
                    break
                md5.update(chunk)
        return md5.hexdigest()

    def get_template_config(
        self, type_: str, template: str, branch: str = 'main'
    ) -> Optional['TemplateConfigModel']:
        """Return the data from the template.yaml file."""
        self.log.info(
            f'action=get-template-config, type={type_}, '
            f'template={template}, cache-valid={self.cache_valid}, branch={branch}'
        )
        if self.cache_valid is True:
            config = self.db_get_config(type_, template)
            if config is not None:
                return config

        url = f'{self.base_raw_url}/{branch}/{type_}/{template}/template.yaml'
        if template == '_app_common':
            type_ = '_app_common'
            url = f'{self.base_raw_url}/{branch}/_app_common/template.yaml'

        params = {}
        if branch:
            params['ref'] = branch
        r: 'Response' = self.session.get(url)
        self.log.debug(f'action=get-template-config, url={url}, status-code={r.status_code}')
        if not r.ok:
            self.log.error(
                f'action=get-template-config, url={r.request.url}, '
                f'status_code={r.status_code}, headers={r.headers}, '
                f'response={r.text or r.reason}'
            )
            self.errors = True
            return None

        try:
            template_config_data = yaml.safe_load(r.text)
            template_config_data.update({'name': template, 'type': type_})
            config = TemplateConfigModel(**template_config_data)

            # upsert db
            self.db_add_config(config)
            self.log.debug(f'action=get-template-config, config={config}')
            return config
        except ValidationError as ex:
            for error in json.loads(ex.json()):
                location = [str(location) for location in error.get('loc')]
                self.log.error(
                    '''Schema validation failed for template.yaml. '''
                    f'''({error.get('msg')}: {' -> '.join(location)})'''
                )
                self.errors = True
            return None

    def get_template_contents(
        self,
        branch: str,
        data: dict,
        path: str,
        type_: str,
        app_builder: bool,
        template_name: str = None,  # preserve the template name for recursion
    ) -> dict:
        """Get the contents of a template and allow recursion."""
        for item in self.contents(branch, type_, path, app_builder):
            # add template name and type to the item to be
            # used during the download and write process
            item['template_name'] = template_name
            item['template_type'] = type_
            # add relative path key AFTER template values have been added to the item
            item['relative_path'] = self.item_relative_path(item)

            # process any files and recurse into any directories
            if item.get('type') == 'file':
                # templates are hierarchical, overwrite previous values with new
                # values using relative path for as the identifier for the file
                data[str(item.get('relative_path'))] = item

                # update manifest data, this will be used during
                # updates to determine if the file has changed
                self.template_manifest.setdefault(item.get('path'), {})
                self.template_manifest[item.get('path')]['sha'] = item.get('sha')
            elif item.get('type') == 'dir':
                nested_dir = f'''{path}/{item.get('name')}'''
                self.get_template_contents(
                    branch, data, nested_dir, type_, app_builder, template_name
                )

        return data

    def init(self, branch: str, type_: str, template: str, app_builder: bool) -> List[dict]:
        """Initialize an App with template files."""
        data = {}
        for tp in self.template_parents(type_, template, branch):
            self.get_template_contents(branch, data, tp, type_, app_builder, tp)
        return data.values()

    def item_relative_path(self, item: dict) -> Path:
        """Return the relative path to the item."""
        path = item.get('path')
        template_path = self.item_template_path(item)

        # handle nested files by stripping the type/template values from the path, the
        # remaining part is the relative path to the file to be used for read/write.
        _path = item.get('name')
        if path.startswith(template_path):
            _path = path.replace(template_path, '')
        return Path(_path)

    @staticmethod
    def item_template_path(item: dict) -> str:
        """Return the template path."""
        if item.get('template_name') == '_app_common':
            return f'''{item.get('template_name')}/'''
        return f'''{item.get('template_type')}/{item.get('template_name')}/'''

    def list(self, branch: str, type_: Optional[str] = None):
        """List template types."""
        template_types = self.template_types
        if type_ is not None:
            if type_ not in self.template_types:
                raise ValueError(f'Invalid Types: {type_}')
            template_types = [type_]

        for selected_type in template_types:
            for td in self.contents(branch, selected_type):
                if td.get('type') == 'dir':
                    template_config = self.get_template_config(
                        selected_type, td.get('name'), branch
                    )
                    if template_config is not None:
                        self.template_data.setdefault(selected_type, [])
                        self.template_data[selected_type].append(template_config)

    def print_error_message(self):
        """Print error message, if applicable."""
        if self.errors is True:
            print(
                '''\n\nErrors were encountered during command execution. Please '''
                f'''see logs at {os.path.join(self.cli_out_path, 'tcex.log')}.'''
            )

    def print_list(self, branch: Optional[str] = None):
        """Print the list output."""
        for type_, templates in self.template_data.items():
            self.print_title(f'''{type_.replace('_', ' ').title()} Templates''')
            for config in templates:
                self.print_setting('Template', config.name, fg_color='green', bold=False)
                self.print_setting('Contributor', config.contributor, fg_color='green', bold=False)
                self.print_setting('Summary', config.summary, fg_color='green', bold=False)
                install_cmd = f'tcex init --type {type_} --template {config.name}'
                if branch != 'main':
                    install_cmd += f' --branch {branch}'
                self.print_setting(
                    'Install Command',
                    install_cmd,
                    fg_color='white',
                    bold=False,
                )
                # looks cleaner just to have a blank line
                # self.print_divider()
                print('')

    @cached_property
    def project_sha(self) -> Optional[str]:
        """Return the current commit sha for the tcex-app-templates project."""
        params = {'perPage': '1'}
        r: 'Response' = self.session.get(f'{self.base_url}/commits', params=params)
        if not r.ok:
            self.log.error(
                f'action=get-project-sha, url={r.request.url}, '
                f'status_code={r.status_code}, headers={r.headers}, '
                f'response={r.text or r.reason}'
            )
            self.errors = True
            return None

        try:
            commits_data = r.json()
        except Exception:
            return None

        # get current sha
        return commits_data[0].get('sha')

    @cached_property
    def session(self) -> 'Session':
        """Return session object"""
        session = Session()
        session.headers.update({'Cache-Control': 'no-cache'})

        # add auth if set (typically not require since default site is public)
        if self.username is not None and self.password is not None:
            session.auth = HTTPBasicAuth(self.username, self.password)

        return session

    @cached_property
    def template_manifest(self) -> dict:
        """Write the template manifest file."""
        if self.template_manifest_fqfn.is_file():
            with self.template_manifest_fqfn.open() as fh:
                return json.load(fh)
        return {}

    def template_manifest_write(self):
        """Write the template manifest file."""
        with self.template_manifest_fqfn.open(mode='w') as fh:
            fh.write(json.dumps(self.template_manifest, indent=2, sort_keys=True))
            fh.write('\n')

    def template_parents(self, type_: str, template: str, branch: str = 'main') -> List[str]:
        """Return all parents for the provided template."""
        # get the config for the requested template
        template_config = self.get_template_config(type_, template, branch)

        # fail if template config can't be found
        if template_config is None:
            self.print_failure(
                f'Failed retrieving template.yaml: type={type_}, template={template}.\n'
                'Try running "tcex list" to get valid template types and names.'
            )

        # iterate over each parent template
        app_templates = []
        for parent in template_config.template_parents:
            parent_config = self.get_template_config(type_, parent, branch)

            # update templates
            app_templates.extend(
                [t for t in parent_config.template_parents if t not in app_templates]
            )

        # add parent after parent->parents have been added
        app_templates.extend(
            [t for t in template_config.template_parents if t not in app_templates]
        )

        # add this current template last
        app_templates.append(template)

        return app_templates

    @property
    def template_to_prefix_map(self) -> Dict[str, str]:
        """Return the defined template types."""
        return {
            'api_service': 'tcva',
            'feed_api_service': 'tcvf',
            'organization': 'tc',
            'playbook': 'tcpb',
            'trigger_service': 'tcvc',
            'web_api_service': 'tcvp',
            'webhook_trigger_service': 'tcvw',
        }

    @property
    def template_types(self) -> List[str]:
        """Return the defined template types."""
        return [
            'api_service',
            'external',
            'feed_api_service',
            'organization',
            'playbook',
            'trigger_service',
            # 'web_api_service',
            'webhook_trigger_service',
        ]

    def update(self, branch: str, template: str, type_: str, ignore_hash=False) -> List[dict]:
        """Initialize an App with template files."""
        # update tcex.json model
        if template is not None:
            self.tj.model.template_name = template
        if type_ is not None:
            self.tj.model.template_type = type_

        # retrieve ALL template contents
        data = {}
        for tp in self.template_parents(
            self.tj.model.template_type, self.tj.model.template_name, branch
        ):
            self.get_template_contents(branch, data, tp, self.tj.model.template_type, False, tp)

        # determine which files should be downloaded
        downloads = []
        for item in data.values():
            # get the relative path to the file
            relative_path = item.get('relative_path')

            # skip files if it has not changed
            if relative_path.is_file() and ignore_hash is False:
                # skip files if it has not changed
                if self.update_item_check_hash(relative_path, item) is False:
                    continue

            # is the file a template file and dev says overwrite
            if self.update_item_prompt(branch, item) and relative_path.is_file() is True:
                response = self.prompt_choice(
                    f'Overwrite: {relative_path}', choices=['y', 'n'], default='n'
                )
                if response == 'n':
                    continue

            downloads.append(item)
            self.template_manifest.setdefault(item.get('path'), {})
            self.template_manifest[item.get('path')]['sha'] = item.get('sha')

        return downloads

    def update_item_check_hash(self, fqfn: Path, item: dict) -> bool:
        """Check if the file hash has changed since init or last update."""
        file_hash = self.file_hash(fqfn)
        if self.template_manifest.get(item.get('path'), {}).get('md5') != file_hash:
            self.log.debug(
                f'''action=update-check-hash, template-file={item.get('name')}, '''
                'check=hash-check, result=hash-has-not-changed'
            )
            return True
        return False

    def update_item_prompt(self, branch: str, item: dict) -> bool:
        """Update the prompt value for the provided item."""
        template_name = item.get('template_name')
        template_config = self.get_template_config(
            self.tj.model.template_type, template_name, branch
        )

        # determine if file requires user prompt
        if str(item.get('relative_path')) in template_config.template_files:
            return False
        return True

    def update_tcex_json(self):
        """Update the tcex.json file."""
        self.tj.model.template_repo_hash = self.project_sha
        self.tj.write()



"""Decorators module for TcEx Framework"""
# flake8: noqa
# first-party
from tcex.decorators.benchmark import Benchmark
from tcex.decorators.debug import Debug
from tcex.decorators.fail_on_output import FailOnOutput
from tcex.decorators.on_exception import OnException
from tcex.decorators.on_success import OnSuccess
from tcex.decorators.output import Output



"""TcEx JSON App Config"""
# standard library
import json
import logging
import os
from collections import OrderedDict
from pathlib import Path
from typing import Optional

# third-party
import colorama as c

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.app_config.models import TcexJsonModel
from tcex.app_config.tcex_json_update import TcexJsonUpdate
from tcex.backports import cached_property

# get tcex logger
tcex_logger = logging.getLogger('tcex')


class TcexJson:
    """Provide a model for the tcex.json config file."""

    def __init__(
        self,
        filename: Optional[str] = None,
        path: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize class properties."""
        filename = filename or 'tcex.json'
        path = path or os.getcwd()
        self.log = logger or tcex_logger

        # properties
        self.fqfn = Path(os.path.join(path, filename))
        self.ij = InstallJson(logger=self.log)

    @cached_property
    def contents(self) -> dict:
        """Return tcex.json file contents."""
        _contents = {}

        if self.fqfn.is_file():
            try:
                with self.fqfn.open() as fh:
                    _contents = json.load(fh, object_pairs_hook=OrderedDict)
            except OSError:  # pragma: no cover
                self.log.error(
                    f'feature=tcex-json, exception=failed-reading-file, filename={self.fqfn}'
                )
        else:  # pragma: no cover
            self.log.error(f'feature=tcex-json, exception=file-not-found, filename={self.fqfn}')

        return _contents

    @cached_property
    def model(self) -> 'TcexJsonModel':
        """Return the Install JSON model."""
        return TcexJsonModel(**self.contents)

    def print_warnings(self):
        """Print warning messages for tcex.json file."""

        # raise error if tcex.json is missing app_name field
        if self.model.package.app_name is None:  # pragma: no cover
            raise RuntimeError('The tcex.json file is missing the package.app_name field.')

        # log warning for old Apps
        if self.model.package.app_version is not None:
            print(
                f'{c.Fore.YELLOW}'
                'WARNING: The tcex.json file defines "app_version" which should only be '
                'defined in legacy Apps. Removing the value can cause the App to be treated '
                'as a new App by TcExchange. Please remove "app_version" when appropriate.'
                f'{c.Fore.RESET}'
            )

    @property
    def update(self) -> 'TcexJsonUpdate':
        """Return InstallJsonUpdate instance."""
        return TcexJsonUpdate(tj=self)

    def write(self):
        """Write current data file."""
        data = self.model.json(exclude_defaults=True, exclude_none=True, indent=2, sort_keys=True)
        with self.fqfn.open(mode='w') as fh:
            fh.write(f'{data}\n')



"""TcEx Utilities String Operations Module"""
# standard library
import random
import re
import string
from functools import reduce
from typing import List, Optional, Union

# first-party
from tcex.backports import cached_property


class StringOperations:
    """TcEx Utilities String Operations Class"""

    @staticmethod
    def camel_string(string_: str) -> 'CamelString':
        """Return custom str with custom properties/methods."""
        return CamelString(string_)

    def camel_to_snake(self, camel_string: str) -> str:
        """Return snake case string from a camel case string.

        Args:
            camel_string: The camel case input string.
        """
        return self._camel_pattern.sub('_', camel_string).lower()

    def camel_to_space(self, camel_string: str) -> str:
        """Return space case string from a camel case string.

        Args:
            camel_string: The camel case input string.
        """
        return self._camel_pattern.sub(' ', camel_string).lower()

    @cached_property
    def inflect(self) -> 'inflect.engine':  # pylint: disable=no-self-use
        """Return instance of inflect."""
        # third-party
        import inflect

        return inflect.engine()

    @staticmethod
    def random_string(string_length: int = 10) -> str:
        """Generate a random string of fixed length

        Args:
            string_length: The length of the string.
        """
        return ''.join(random.choice(string.ascii_letters) for _ in range(string_length))  # nosec

    @staticmethod
    def snake_string(string_: str) -> 'SnakeString':
        """Return custom str with custom properties/methods."""
        return SnakeString(string_)

    @staticmethod
    def snake_to_pascal(snake_string: str) -> str:
        """Convert snake_case to PascalCase

        Args:
            snake_string: The snake case input string.
        """
        return snake_string.replace('_', ' ').title().replace(' ', '')

    @staticmethod
    def snake_to_camel(snake_string: str) -> str:
        """Convert snake_case to camelCase

        Args:
            snake_string: The snake case input string.
        """
        components = snake_string.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])

    @staticmethod
    def to_bool(value: Union[bool, int, str]) -> bool:
        """Convert value to bool.

        Args:
            value: The value to convert to boolean.
        """
        return str(value).lower() in ['1', 't', 'true', 'y', 'yes']

    @staticmethod
    def truncate_string(
        t_string: str,
        length: int,
        append_chars: Optional[str] = '',
        spaces: Optional[bool] = False,
    ) -> str:
        """Truncate a string to a given length.

        Args:
            t_string: The input string to truncate.
            length: The length of the truncated string.
            append_chars: Any character that should be appended to the
                string. Typically used for ellipsis (e.g. ...).
            spaces: If True truncation will be done at the
                nearest space before the truncation length to avoid chopping words.
        """
        if t_string is None:
            t_string = ''

        if length is None:
            length = len(t_string)

        if len(t_string) <= length:
            return t_string

        # set sane default for append_chars
        append_chars = str(append_chars or '')

        # ensure append_chars is not longer than length
        if len(append_chars) > length:  # pragma: no cover
            raise RuntimeError('Append chars cannot exceed the truncation length.')

        output = t_string[0 : length - len(append_chars)]  # noqa: E203
        if spaces is True:
            if not output.endswith(' '):
                # split output on spaces and drop last item to terminate string on word
                output = ' '.join(output.split(' ')[:-1])

        return f'{output.rstrip()}{append_chars}'

    @property
    def _camel_pattern(self) -> 're.Pattern':
        """Return compiled re pattern."""
        return re.compile(r'(?<!^)(?=[A-Z])')

    @staticmethod
    def wrap_string(
        line: str,
        wrap_chars: Optional[List[str]] = None,
        length: int = 100,
        force_wrap=True,
    ) -> str:
        """Wrap a long string to a given length.

        Lines will only be broken on instances of strings from wrap_chars.

        Params:
            line - the string to break into lines
            wrap_chars - list of strings line should be broken on
            length - max length for any single line
            force_wrap - if True, line will be broken even if no string from wrap_chars is
                available
        """
        wrap_chars = wrap_chars or [' ']

        # if line is already shorter than max length, we're all good!
        if len(line) < length:
            return line

        def _tokenize(acc, curr):
            """Tokenize the input into strings and separators (from wrap_chars)."""
            if curr in wrap_chars:
                return acc + [curr] + ['']
            if len(acc[-1]) == length and force_wrap:
                return acc + [curr]
            acc[-1] += curr
            return acc

        tokens = reduce(_tokenize, line, [''])

        def _chop(acc, curr):
            """Build strings in acc such that no string is greater than length chars.

            Breaks at chars in wrap_chars.
            """
            if len(acc[-1]) + len(curr) < length:
                acc[-1] += curr
                return acc
            return acc + [curr]

        lines = reduce(_chop, tokens, [''])

        return '\n'.join(lines).strip()


class CamelString(str):
    """Power String"""

    # properties
    so = StringOperations()

    def pascal_case(self):
        """Return camel case version of string."""
        return CamelString(self.so.camel_to_space(self).title().replace(' ', ''))

    def plural(self):
        """Return inflect.plural spelling of string."""
        return CamelString(self.so.inflect.plural(self.singular()))

    def singular(self):
        """Return inflect.singular spelling of string."""
        try:
            _singular = self.so.inflect.singular(self)
        except Exception:
            _singular = self.so.inflect.singular_noun(self)

        if not _singular:
            _singular = self
        return CamelString(_singular)

    def snake_case(self):
        """Return camel case version of string."""
        return CamelString(self.so.camel_to_snake(self))

    def space_case(self):
        """Return string snake to camel."""
        return CamelString(self.so.camel_to_space(self))  # type: ignore


class SnakeString(str):
    """Power String"""

    # properties
    so = StringOperations()

    def camel_case(self):
        """Return camel case version of string."""
        return SnakeString(self.so.snake_to_camel(self))

    def pascal_case(self):
        """Return camel case version of string."""
        return SnakeString(self.so.snake_to_pascal(self))

    def plural(self):
        """Return inflect.plural spelling of string."""
        return SnakeString(self.so.inflect.plural(self.singular()))

    def singular(self):
        """Return inflect.singular spelling of string."""
        try:
            _singular = self.so.inflect.singular(self)
        except Exception:
            _singular = self.so.inflect.singular_noun(self)

        if not _singular:
            _singular = self
        return SnakeString(_singular)

    def space_case(self):
        """Return a space case version of a string

        _ will be replaced with spaces.
        """
        return self.replace('_', ' ').title()



"""Artifact_Type / Artifact_Types Model"""
# pylint: disable=no-member,no-self-argument,no-self-use,wrong-import-position
# standard library
from typing import List, Optional

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.utils import Utils


class ArtifactTypesModel(
    BaseModel,
    title='ArtifactTypes Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Artifact_Types Model"""

    _mode_support = PrivateAttr(False)

    data: Optional[List['ArtifactTypeModel']] = Field(
        [],
        description='The data for the ArtifactTypes.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


class ArtifactTypeDataModel(
    BaseModel,
    title='ArtifactType Data Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Artifact_Types Data Model"""

    data: Optional[List['ArtifactTypeModel']] = Field(
        [],
        description='The data for the ArtifactTypes.',
        methods=['POST', 'PUT'],
        title='data',
    )


class ArtifactTypeModel(
    V3ModelABC,
    alias_generator=Utils().snake_to_camel,
    extra=Extra.allow,
    title='ArtifactType Model',
    validate_assignment=True,
):
    """Artifact_Type Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(True)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    data_type: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The **data type** for the Artifact_Type.',
        read_only=True,
        title='dataType',
    )
    derived_link: bool = Field(
        None,
        allow_mutation=False,
        description='The **derived link** for the Artifact_Type.',
        read_only=True,
        title='derivedLink',
    )
    description: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The **description** for the Artifact_Type.',
        read_only=True,
        title='description',
    )
    id: Optional[int] = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    intel_type: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The **intel type** for the Artifact_Type.',
        read_only=True,
        title='intelType',
    )
    name: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The **name** for the Artifact_Type.',
        read_only=True,
        title='name',
    )


# add forward references
ArtifactTypeDataModel.update_forward_refs()
ArtifactTypeModel.update_forward_refs()
ArtifactTypesModel.update_forward_refs()



"""TcEx JSON Update"""
# standard library
import os
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:  # pragma: no cover
    from .tcex_json import TcexJson


class TcexJsonUpdate:
    """Update install.json file with current standards and schema."""

    def __init__(self, tj: 'TcexJson'):  # pylint: disable=E0601
        """Initialize class properties."""
        self.tj = tj

    def multiple(self, template: Optional[str] = None):
        """Update the contents of the tcex.json file."""

        # update app_name
        self.update_package_app_name()

        # update deprecated fields
        # self.update_deprecated_fields()

        # update package excludes
        self.update_package_excludes()

        # update package excludes
        self.update_lib_versions()

        # update template
        if template is not None:
            self.tj.template = template

        # write updated profile
        self.tj.write()

    def update_package_app_name(self):
        """Update the package app_name in the tcex.json file."""
        if (
            self.tj.model.package.app_name is None
            or self.tj.model.package.app_name in self.tj.ij.app_prefixes.values()
        ):
            # lower case name and replace prefix if already exists
            _app_name = (
                os.path.basename(os.getcwd()).lower().replace(self.tj.ij.app_prefix.lower(), '')
            )

            # replace spaces and dashes with underscores
            _app_name = _app_name.replace(' ', '_').replace('-', '_').lower()

            # title case app name
            _app_name = '_'.join([a.title() for a in _app_name.split('_')])

            # prepend appropriate App prefix (e.g., TCPB_-_)
            _app_name = f'{self.tj.ij.app_prefix}{_app_name}'

            # update App name
            self.tj.model.package.app_name = _app_name

    # def update_deprecated_fields(self):
    #     """Update deprecated fields in the tcex.json file."""
    #     deprecated_fields = ['profile_include_dirs']
    #     for d in deprecated_fields:
    #         setattr(self.tj.model, d, None)

    def update_package_excludes(self):
        """Update the excludes values in the tcex.json file."""
        for i in [
            '.gitignore',
            '.pre-commit-config.yaml',
            'local-*',
            'pyproject.toml',
            'setup.cfg',
            'tcex.json',
        ]:
            if i not in self.tj.model.package.excludes:
                # TODO: [low] pydantic doesn't seem to allow removing items from list???
                self.tj.model.package.excludes.append(i)

    def update_lib_versions(self):
        """Update the lib_versions array in the tcex.json file."""
        if os.getenv('TCEX_LIB_VERSIONS') and not self.tj.model.lib_versions:
            _lib_versions = []
            for version in os.getenv('TCEX_LIB_VERSIONS').split(','):
                _lib_versions.append(
                    {
                        'lib_dir': f'lib_${{env:{version}}}',
                        'python_executable': f'~/.pyenv/versions/${{env:{version}}}/bin/python',
                    }
                )
            self.tj.model.lib_versions = _lib_versions



"""Tag / Tags Model"""
# pylint: disable=no-member,no-self-argument,no-self-use,wrong-import-position
# standard library
from datetime import datetime
from typing import List, Optional

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.utils import Utils


class TagsModel(
    BaseModel,
    title='Tags Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Tags Model"""

    _mode_support = PrivateAttr(True)

    data: Optional[List['TagModel']] = Field(
        [],
        description='The data for the Tags.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


class TagDataModel(
    BaseModel,
    title='Tag Data Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Tags Data Model"""

    data: Optional[List['TagModel']] = Field(
        [],
        description='The data for the Tags.',
        methods=['POST', 'PUT'],
        title='data',
    )


class TagModel(
    V3ModelABC,
    alias_generator=Utils().snake_to_camel,
    extra=Extra.allow,
    title='Tag Model',
    validate_assignment=True,
):
    """Tag Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(True)
    _staged = PrivateAttr(False)

    cases: Optional['CasesModel'] = Field(
        None,
        allow_mutation=False,
        description='The **cases** for the Tag.',
        read_only=True,
        title='cases',
    )
    description: Optional[str] = Field(
        None,
        description='A brief description of the Tag.',
        methods=['POST', 'PUT'],
        max_length=255,
        min_length=0,
        read_only=False,
        title='description',
    )
    groups: Optional['GroupsModel'] = Field(
        None,
        allow_mutation=False,
        description='The **groups** for the Tag.',
        read_only=True,
        title='groups',
    )
    id: Optional[int] = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    indicators: Optional['IndicatorsModel'] = Field(
        None,
        allow_mutation=False,
        description='The **indicators** for the Tag.',
        read_only=True,
        title='indicators',
    )
    last_used: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Tag was last used.',
        read_only=True,
        title='lastUsed',
    )
    name: Optional[str] = Field(
        None,
        description='The **name** for the Tag.',
        methods=['POST', 'PUT'],
        max_length=128,
        min_length=1,
        read_only=False,
        title='name',
    )
    owner: Optional[str] = Field(
        None,
        description='The name of the Owner of the Tag.',
        methods=['POST'],
        read_only=False,
        title='owner',
    )
    victims: Optional['VictimsModel'] = Field(
        None,
        allow_mutation=False,
        description='The **victims** for the Tag.',
        read_only=True,
        title='victims',
    )

    @validator('cases', always=True)
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()
        return v

    @validator('groups', always=True)
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()
        return v

    @validator('indicators', always=True)
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()
        return v

    @validator('victims', always=True)
    def _validate_victims(cls, v):
        if not v:
            return VictimsModel()
        return v


# first-party
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorsModel
from tcex.api.tc.v3.victims.victim_model import VictimsModel

# add forward references
TagDataModel.update_forward_refs()
TagModel.update_forward_refs()
TagsModel.update_forward_refs()



"""Playbook ABC"""
# standard library
import base64
import json
import logging
import re
from collections import OrderedDict
from typing import Any, List, Optional, Union

# first-party
from tcex.key_value_store import KeyValueApi, KeyValueRedis
from tcex.pleb.registry import registry
from tcex.utils.utils import Utils
from tcex.utils.variables import BinaryVariable, StringVariable

# get tcex logger
logger = logging.getLogger('tcex')


class PlaybookRead:
    """Playbook Read

    Args:
        key_value_store: A KV store instance.
        context: The KV Store context/session_id. For PB Apps the context is provided on
            startup, but for service Apps each request gets a different context.
        output_variables: The requested output variables. For PB Apps outputs are provided on
            startup, but for service Apps each request gets different outputs.
    """

    def __init__(
        self,
        context: str,
        key_value_store: Union[KeyValueApi, KeyValueRedis],
    ):
        """Initialize the class properties."""
        self.context = context
        self.key_value_store = key_value_store

        # properties
        self.log = logger
        self.utils = Utils()

    def _check_variable_type(self, variable: str, type_: str) -> bool:
        """Validate the correct type was passed to the method."""
        if self.utils.get_playbook_variable_type(variable).lower() != type_.lower():
            raise RuntimeError(
                f'Invalid variable provided ({variable}), variable must be of type {type_}.'
            )

    @staticmethod
    def _coerce_string_value(value: Union[bool, float, int, str]) -> str:
        """Return a string value from an bool or int."""
        # coerce bool before int as python says a bool is an int
        if isinstance(value, bool):
            # coerce bool to str type
            value = str(value).lower()

        # coerce int to str type
        if isinstance(value, (float, int)):
            value = str(value)

        return value

    @staticmethod
    def _decode_binary(data: bytes) -> str:
        """Return decoded bytes data handling data written by java apps."""
        try:
            data = data.decode('utf-8')
        except UnicodeDecodeError:  # pragma: no cover
            # for data written an upstream java App
            data = data.decode('latin-1')
        return data

    def _get_data(self, key: str) -> Any:
        """Get the value from Redis if applicable."""
        value = None
        try:
            value = self.key_value_store.read(self.context, key.strip())
        except RuntimeError as e:
            self.log.error(e)
        return value

    @staticmethod
    def _load_data(value: str) -> dict:
        """Return the loaded JSON value or raise an error."""
        try:
            return json.loads(value, object_pairs_hook=OrderedDict)
        except ValueError as e:  # pragma: no cover
            raise RuntimeError(f'Failed to JSON load data "{value}" ({e}).')

    def _null_key_check(self, key: Any) -> bool:
        """Return False if value is not null."""
        if key is None:
            self.log.warning('The provided key was None.')
            return True

        return False

    def _process_binary(
        self, data: str, b64decode: bool, decode: bool, serialized: bool
    ) -> Optional[Union[str, bytes]]:
        """Process the provided."""
        if data is not None:
            # Single type are serialized, array types are not
            if serialized is True:
                data = self._load_data(data)

            if b64decode is True:
                data = BinaryVariable(base64.b64decode(data))
                if decode is True:
                    # allow developer to decided if they want bytes or str
                    data = self._decode_binary(data)
            elif isinstance(data, bytes):
                # data is likely returned base64 encoded bytes string, the App
                # should get the base64 as a str if b64decode is False
                data = data.decode()

        return data

    def _process_key_value(
        self, data: str, resolve_embedded: bool, serialized: bool
    ) -> Optional[dict]:
        """Read the value from key value store.

        KeyValue data should be stored as a JSON string.
        """
        if data is not None:
            # decode in case data comes back from kvstore as bytes
            if isinstance(data, bytes):
                data = data.decode()

            # Single type are serialized, array types are not
            if serialized is True:
                data = self._load_data(data)

            # IMPORTANT:
            # A Single level of nested variables is supported. There is no way
            # in the TC platform to create double nested variables. Any App
            # that would try and create a double nested variable is improperly
            # written.

            # KeyValue List Input
            # -------------------------------------------------------------
            # | key                         | value                       |
            # =============================================================
            # | my_binary                   | #App:7979:two!Binary        |
            # -------------------------------------------------------------
            # | my_binary_array             | #App:7979:two!BinaryArray   |
            # -------------------------------------------------------------
            # | my_key_value                | #App:7979:two!KeyValue      |
            # -------------------------------------------------------------
            # | my_key_value_array          | #App:7979:two!KeyValueArray |
            # -------------------------------------------------------------
            # | my_string                   | #App:7979:two!String        |
            # -------------------------------------------------------------
            # | my_string_array             | #App:7979:two!StringArray   |
            # -------------------------------------------------------------
            # | my_tcentity                 | #App:7979:two!TCEntity      |
            # -------------------------------------------------------------
            # | my_tcentity_array           | #App:7979:two!TCEntityArray |
            # -------------------------------------------------------------

            # An upstream Apps KeyValue output can be used in a KeyValueList input, but
            # Apps SHOULD NOT be writing KeyValueArray with nested variables. This means
            # that there will only ever be 1 levels of nesting.

            # KeyValueList Input -> Nested KeyValue/KeyValueArray, the nested
            # KeyValue/KeyValueArray CAN NOT have nested variables since they
            # have to come from an upstream App.

            # check if keyvalue value is a variable
            if resolve_embedded:
                value = data.get('value')
                if self.utils.is_playbook_variable(value):
                    # any type can be nested, but no further nesting is supported
                    data['value'] = self.any(value)
                else:
                    # read embedded is less efficient and has more caveats
                    data['value'] = self._read_embedded(value)

        return data

    @staticmethod
    def _process_space_patterns(string: str) -> str:
        r"""Return the string with \s replace with spaces."""
        # replace "\s" with a space only for user input.
        # using '\\s' will prevent replacement.
        string = re.sub(r'(?<!\\)\\s', ' ', string)
        string = re.sub(r'\\\\s', r'\\s', string)
        return string

    def _process_string(self, data: str, resolve_embedded: bool, serialized: bool) -> Optional[str]:
        """Process the provided."""
        if data is not None:
            # decode in case data comes back from kvstore as bytes
            if isinstance(data, bytes):
                data = data.decode()

            # Single type are serialized, array types are not
            if serialized is True:
                data = self._load_data(data)

            # only resolve embedded variables if resolve_embedded is True and
            # the entire string does not exactly match a variable pattern
            if resolve_embedded and not self.utils.is_playbook_variable(data):
                data = self._read_embedded(data)

            # coerce data back to string, since technically TC doesn't support bool, int, etc
            data = StringVariable(self._coerce_string_value(data))

        return data

    def _process_tc_batch(self, data: str, serialized: bool) -> Optional[dict]:
        """Process the provided."""
        if data is not None:
            # decode in case data comes back from kvstore as bytes
            if isinstance(data, bytes):
                data = data.decode()

            # Single type are serialized, array types are not
            if serialized is True:
                data = self._load_data(data)

        return data

    def _process_tc_entity(self, data: str, serialized: bool) -> Optional[dict]:
        """Process the provided."""
        if data is not None:
            # decode in case data comes back from kvstore as bytes
            if isinstance(data, bytes):
                data = data.decode()

            # Single type are serialized, array types are not
            if serialized is True:
                data = self._load_data(data)

        return data

    def _read_embedded(self, value: str) -> str:
        """Read method for "embedded" variables.

        .. Note:: The ``read()`` method will automatically determine if the input is a variable or
            needs to be searched for embedded variables.

        Embedded variable rules:

        * Only user input can have embedded variables.
        * Only String and KeyValueArray variables can have embedded variables.
        * Variables can only be embedded one level deep.

        This method will automatically convert variables embedded in a string with value retrieved
        from DB. If there are no keys/variables the raw string will be returned.

        Examples::

            DB Values
            #App:7979:variable_name!String:
                "embedded \\"variable\\""
            #App:7979:two!String:
                "two"
            #App:7979:variable_name!StringArray:
                ["one", "two", "three"]

            Examples 1:
                Input:  "This input has a embedded #App:7979:variable_name!String"

            Examples 2:
                Input: ["one", #App:7979:two!String, "three"]

            Examples 3:
                Input: [{
                    "key": "embedded string",
                    "value": "This input has a embedded #App:7979:variable_name!String"
                }, {
                    "key": "string array",
                    "value": #App:7979:variable_name!StringArray
                }, {
                    "key": "string",
                    "value": #App:7979:variable_name!String
                }]

        Args:
            value (str): The value to parsed and updated from the DB.

        Returns:
            (str): Results retrieved from DB
        """
        if value is None:  # pragma: no cover
            return value

        for match in re.finditer(self.utils.variable_expansion_pattern, str(value)):
            variable = match.group(0)  # the full variable pattern
            if match.group('origin') == '#':  # pb-variable
                v = self.any(variable)
            elif match.group('origin') == '&':  # tc-variable
                v = registry.resolve_variable(variable)

            # TODO: [high] should this behavior be changed in 3.0?
            self.log.trace(f'embedded variable: {variable}, value: {v}')

            if match.group('type') in ['Binary', 'BinaryArray']:
                self.log.trace(
                    f'Binary types may not be embedded into strings. Could not embed: {variable}'
                )
                v = '<binary>'

            if isinstance(v, (dict, list)):
                v = json.dumps(v)

            elif v is None:
                v = '<null>'

            # value.replace was chosen over re.sub due to an issue encountered while testing an app.
            # re.sub will handle escaped characters like \t, value.replace would not handle these
            # scenarios.
            value = value.replace(variable, v)

        return StringVariable(value)

    @staticmethod
    def _to_array(value: Optional[Union[List, str]]) -> List:
        """Return the provided array as a list."""
        if value is None:
            # Adding none value to list breaks App logic. It's better to not request
            # Array and build array externally if None values are required.
            value = []
        elif not isinstance(value, list):
            value = [value]
        return value

    def any(self, key: str) -> Optional[Union[bytes, dict, list, str]]:
        """Return the value from the keystore for all types.

        This is a quick helper method, for more advanced features
        the individual read methods should be used (e.g., binary).
        """
        if self._null_key_check(key) is True:
            return None

        key = key.strip()  # clean up key
        variable_type = self.utils.get_playbook_variable_type(key).lower()
        variable_type_map = {
            'binary': self.binary,
            'binaryarray': self.binary_array,
            'keyvalue': self.key_value,
            'keyvaluearray': self.key_value_array,
            'string': self.string,
            'stringarray': self.string_array,
            'tcbatch': self.tc_batch,
            'tcentity': self.tc_entity,
            'tcentityarray': self.tc_entity_array,
            # 'tcenhancedentity': self.tc_entity,
        }
        value = variable_type_map.get(variable_type, self.raw)(key)

        if value is not None:
            if variable_type == 'binary':
                value = BinaryVariable(value)
            elif variable_type == 'binaryarray':
                value = [v if v is None else BinaryVariable(v) for v in value]
            elif variable_type == 'string':
                value = StringVariable(value)
            elif variable_type == 'stringarray':
                value = [v if v is None else StringVariable(v) for v in value]

        return value

    def binary(
        self,
        key: str,
        b64decode: Optional[bool] = True,
        decode: Optional[bool] = False,
    ) -> Optional[Union[str, bytes]]:
        """Read the value from key value store.

        Binary data should be stored as base64 encoded string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'Binary')

        data: Optional[str] = self._get_data(key)
        return self._process_binary(data, b64decode=b64decode, decode=decode, serialized=True)

    def binary_array(
        self,
        key: str,
        b64decode: Optional[bool] = True,
        decode: Optional[bool] = False,
    ) -> Optional[List[Union[bytes, str]]]:
        """Read the value from key value store.

        BinaryArray data should be stored as base64 encoded serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'BinaryArray')

        data: Optional[str] = self._get_data(key)
        if data is not None:
            # data should be base64 encoded bytes string

            # decode the entire response, but not the items in the array?
            if isinstance(data, bytes):
                data = data.decode()

            # Array type is serialized before writing to redis, deserialize the data
            data = self._load_data(data)

            values = []
            for d in data:
                # d should be a base64 encoded string
                values.append(
                    self._process_binary(d, b64decode=b64decode, decode=decode, serialized=False)
                )
            data = values

        return data

    def key_value(
        self,
        key: str,
        resolve_embedded: Optional[bool] = True,
    ) -> Optional[dict]:
        """Read the value from key value store.

        KeyValue data should be stored as a JSON string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'KeyValue')

        data: Optional[str] = self._get_data(key)
        return self._process_key_value(data, resolve_embedded=resolve_embedded, serialized=True)

    def key_value_array(
        self,
        key: str,
        resolve_embedded: Optional[bool] = True,
    ) -> Optional[List[str]]:
        """Read the value from key value store.

        KeyValueArray data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'KeyValueArray')

        data: Optional[str] = self._get_data(key)
        if data is not None:
            # data should be string

            # decode the entire response, but not the items in the array?z
            if isinstance(data, bytes):
                data = data.decode()

            # Array type is serialized before writing to redis, deserialize the data
            data = self._load_data(data)

            values = []
            for d in data:
                # d should be a base64 encoded string
                values.append(
                    self._process_key_value(d, resolve_embedded=resolve_embedded, serialized=False)
                )
            data = values

        return data

    def raw(self, key: str) -> Optional[any]:
        """Read method of CRUD operation for raw data.

        Bytes input will be returned a as string as there is no way
        to determine data from redis originated as bytes or string.
        """
        if self._null_key_check(key) is True:
            return None

        return self.key_value_store.read(self.context, key.strip())

    def string(
        self,
        key: str,
        resolve_embedded: Optional[bool] = True,
    ) -> Optional[str]:
        """Read the value from key value store.

        String data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'String')

        data: Optional[str] = self._get_data(key)
        return self._process_string(data, resolve_embedded=resolve_embedded, serialized=True)

    def string_array(self, key: str) -> Optional[List[str]]:
        """Read the value from key value store.

        StringArray data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'StringArray')

        data: Optional[str] = self._get_data(key)
        if data is not None:
            # data should be string

            # decode the entire response
            if isinstance(data, bytes):
                data = data.decode()

            # Array type is serialized before writing to redis, deserialize the data
            data = self._load_data(data)

            values = []
            for d in data:
                values.append(self._process_string(d, resolve_embedded=False, serialized=False))
            data = values

        return data

    def tc_batch(self, key: str) -> Optional[dict]:
        """Read the value from key value store.

        TCBatch data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'TCBatch')

        data: Optional[str] = self._get_data(key)
        return self._process_tc_batch(data, serialized=True)

    def tc_entity(self, key: str) -> Optional[dict]:
        """Read the value from key value store.

        TCEntity data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'TCEntity')

        data: Optional[str] = self._get_data(key)
        return self._process_tc_entity(data, serialized=True)

    def tc_entity_array(
        self,
        key: str,
    ) -> Optional[List[str]]:
        """Read the value from key value store.

        TCEntityArray data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'TCEntityArray')

        data: Optional[str] = self._get_data(key)
        if data is not None:
            # data should be string

            # decode the entire response
            if isinstance(data, bytes):
                data = data.decode()

            # Array type is serialized before writing to redis, deserialize the data
            data = self._load_data(data)

            values = []
            for d in data:
                # d should be a base64 encoded string
                values.append(self._process_tc_entity(d, serialized=False))
            data = values

        return data

    def variable(
        self, key: str, array: Optional[bool] = False
    ) -> Optional[Union[bytes, dict, list, str]]:
        """Read method of CRUD operation for working with KeyValue DB.

        This method will automatically check to see if a single variable is passed
        or if "mixed" data is passed and return the results from the DB. It will also
        automatically determine the variable type to read.
        """
        value = key
        if isinstance(key, str):
            key = key.strip()

            if re.match(self.utils.variable_playbook_match, key):
                value = self.any(key=key)
            else:
                # replace space patterns
                value = self._process_space_patterns(value)

                # key must be an embedded variable
                value = self._read_embedded(value)

        if array is True:
            value = self._to_array(value)

        return value



"""ThreatConnect HMAC Authorization"""
# standard library
import time
from typing import TYPE_CHECKING, Callable, Union

# third-party
from requests import auth

# first-party
from tcex.input.field_types.sensitive import Sensitive

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from requests import request


class TokenAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, tc_token: Union[Callable, str, 'Sensitive']):
        """Initialize the Class properties."""
        # super().__init__()
        auth.AuthBase.__init__(self)
        self.tc_token = tc_token

    def _token_header(self):
        """Return HMAC Authorization header value."""
        _token = None
        if hasattr(self.tc_token, 'token'):
            # Token Module - The token module is provided that will handle authentication.
            _token = self.tc_token.token.value
        elif callable(self.tc_token):
            # Callabe - A callable method is provided that will return the token as a plain
            #     string. The callable will have to handle token renewal.
            _token = self.tc_token()
        elif isinstance(self.tc_token, Sensitive):
            # Sensitive - A sensitive string type was passed. Likely no support for renewal.
            _token = self.tc_token.value
        else:
            # String - A string type was passed. Likely no support for renewal.
            _token = self.tc_token

        # Return formatted token
        return f'TC-Token {_token}'

    def __call__(self, r: 'request') -> 'request':
        """Add the authorization headers to the request."""
        timestamp = int(time.time())

        # Add required headers to auth.
        r.headers['Authorization'] = self._token_header()
        r.headers['Timestamp'] = timestamp
        return r



"""TcEx Utilities Request to Curl Module"""
# standard library
import re
from typing import TYPE_CHECKING, List, Optional, Union
from urllib.parse import urlsplit

# first-party
from tcex.utils.utils import Utils

if TYPE_CHECKING:
    # third-party
    from requests import Request


class RequestsToCurl:
    """TcEx Utilities Request to Curl Class"""

    def __init__(self):
        """Initialize the Class properties."""
        self.utils = Utils()

    def convert(
        self,
        request: 'Request',
        mask_headers: Optional[bool] = True,
        mask_patterns: List[str] = None,
        **kwargs: Union[bool, dict],
    ) -> str:
        """Return converted Prepared Request to a curl command.

        Args:
            request: The response.request object.
            mask_headers: If True then values for certain header key will be masked.
            mask_patterns: A list of patterns if found in headers the value will be masked.
            body_limit: The size limit for the body value.
            mask_body: If True the body will be masked.
            proxies: A dict containing the proxy configuration.
            verify: If False the curl command will include --insecure flag.
            write_file: If True and the body is binary it will be written as a temp file.
        """
        body_limit = kwargs.get('body_limit', 100)
        proxies = kwargs.get('proxies', {})
        verify = kwargs.get('verify', True)
        # write_file: bool = kwargs.get('write_file', False)

        # APP-79 - adding the ability to log request as curl commands
        cmd = ['curl', '-X', request.method]

        # add headers to curl command
        for k, v in sorted(list(dict(request.headers).items())):
            if mask_headers is True:
                patterns = [
                    'authorization',
                    'cookie',
                    'password',
                    'secret',
                    'session',
                    'token',
                    'username',
                ]
                if isinstance(mask_patterns, list):
                    # add user defined mask patterns
                    patterns.extend(mask_patterns)

                for p in patterns:
                    if re.match(rf'.*{p}.*', k, re.IGNORECASE):
                        v: str = self.utils.printable_cred(v)

                # using gzip in Accept-Encoding with CURL on the CLI produces
                # the warning "Binary output can mess up your terminal."
                if k.lower() == 'accept-encoding':
                    encodings = [e.strip() for e in v.split(',')]
                    for encoding in list(encodings):
                        if encoding in ['gzip']:
                            encodings.remove(encoding)
                    v: str = ', '.join(encodings)

            cmd.append(f"-H '{k}: {v}'")

        if request.body:
            # add body to the curl command
            body = request.body
            try:
                if isinstance(body, bytes):
                    body = body.decode('utf-8')

                if kwargs.get('mask_body', False):
                    # mask_body
                    body = self.utils.printable_cred(body)
                else:
                    # truncate body
                    body = self.utils.truncate_string(
                        t_string=body, length=body_limit, append_chars='...'
                    )
                body_data = f'-d "{body}"'
            except Exception:
                # set static filename so that when running a large job App thousands of files do
                # no get created.
                body_data = '--data-binary @/tmp/body-file'
                # TODO: [super-low] - this is only useful for local testing.
                # if write_file is True:
                #     temp_file: str = self.utils.write_temp_binary_file(
                #         content=body, filename='curl-body'
                #     )
                #     body_data = f'--data-binary @{temp_file}'
            cmd.append(body_data)

        if proxies is not None and proxies.get('https'):
            # parse formatted string {'https': 'bob:pass@https://localhost:4242'}
            proxy_url: Optional[str] = proxies.get('https')
            proxy_data = urlsplit(proxy_url)

            # auth
            if proxy_data.username:
                cmd.extend(['--proxy-user', f'{proxy_data.username}:xxxxx'])

            # server
            proxy_server = proxy_data.hostname
            if proxy_data.port:
                proxy_server = f'{proxy_data.hostname}:{proxy_data.port}'
            cmd.extend(['--proxy', proxy_server])

        if not verify:
            # add insecure flag to curl command
            cmd.append('--insecure')

        # add url to curl command
        cmd.append(request.url)

        return ' '.join(cmd)



#!/usr/bin/env python
"""TcEx Dependencies Command"""
# standard library
import os
import platform
import shutil
import subprocess  # nosec
import sys
from distutils.version import StrictVersion  # pylint: disable=no-name-in-module
from pathlib import Path
from typing import List, Optional
from urllib.parse import quote

# third-party
import typer

# first-party
from tcex.app_config.models.tcex_json_model import LibVersionModel
from tcex.backports import cached_property
from tcex.bin.bin_abc import BinABC


class Dep(BinABC):
    """Install dependencies for App."""

    def __init__(
        self,
        branch: str,
        dev: bool,
        no_cache_dir: bool,
        pre: bool,
        proxy_host: str,
        proxy_port: int,
        proxy_user: str,
        proxy_pass: str,
    ):
        """Initialize Class properties."""
        super().__init__()
        self.branch = branch
        self.dev = dev
        self.no_cache_dir = no_cache_dir
        self.pre = pre
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_user = proxy_user
        self.proxy_pass = proxy_pass

        # properties
        self.env = self._env
        self.latest_version = None
        self.lib_directory = (
            f'lib_{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}'
        )
        self.proxy_enabled = False
        self.requirements_fqfn_branch = None
        self.static_lib_dir = 'lib_latest'

        # update tcex.json
        self.tj.update.multiple()

    def _build_command(self, python_executable: Path, lib_dir: Path) -> List[str]:
        """Build the pip command for installing dependencies.

        Args:
            python_executable: The fully qualified path of the Python executable.
            lib_dir: The fully qualified path of the lib directory.

        Returns:
            list: The Python pip command with all required args.
        """
        # support temp (branch) requirements.txt file
        _requirements_fqfn = str(self.requirements_fqfn)
        if self.requirements_fqfn_branch:
            _requirements_fqfn = str(self.requirements_fqfn_branch)

        exe_command = [
            str(python_executable),
            '-m',
            'pip',
            'install',
            '-r',
            _requirements_fqfn,
            '--ignore-installed',
            '--quiet',
            '--target',
            lib_dir.name,
        ]
        if self.no_cache_dir:
            exe_command.append('--no-cache-dir')
        if self.pre:
            exe_command.append('--pre')
        if self.proxy_enabled:
            # trust the pypi hosts to avoid ssl errors
            trusted_hosts = ['pypi.org', 'pypi.python.org', 'files.pythonhosted.org']

            for host in trusted_hosts:
                exe_command.append('--trusted-host')
                exe_command.append(host)

        return exe_command

    def _create_lib_latest(self):
        """Create the lib_latest symlink for App Builder."""
        if platform.system() == 'Windows':
            shutil.copytree(f'lib_{self.latest_version}', self.static_lib_dir)
        else:
            if os.path.islink(self.static_lib_dir):
                os.unlink(self.static_lib_dir)
            elif os.path.isfile(self.static_lib_dir):
                os.rmdir(self.static_lib_dir)
            os.symlink(f'lib_{self.latest_version}', self.static_lib_dir)

    @property
    def _env(self):
        """Return the environment variables."""
        _env = os.environ.copy()

        # add ci env
        ci_token = os.getenv('CI_JOB_TOKEN')
        if ci_token:
            _env.update({'CI_JOB_TOKEN': ci_token})

        return _env

    @staticmethod
    def _remove_previous(fqpn: Path):
        """Remove previous lib directory recursively."""
        if os.access(fqpn, os.W_OK):
            shutil.rmtree(fqpn)

    def configure_proxy(self):
        """Configure proxy settings using environment variables."""
        if os.getenv('HTTP_PROXY') or os.getenv('HTTPS_PROXY'):
            # don't change proxy settings if the OS already has them configured.
            return

        if self.proxy_host is not None and self.proxy_port is not None:
            # proxy url without auth
            proxy_url = f'{self.proxy_host}:{self.proxy_port}'
            if self.proxy_user is not None and self.proxy_pass is not None:
                proxy_user = quote(self.proxy_user, safe='~')
                proxy_pass = quote(self.proxy_pass, safe='~')

                # proxy url with auth
                proxy_url = f'{proxy_user}:{proxy_pass}@{proxy_url}'

            # update proxy properties
            self.proxy_enabled = True
            self.env.update(
                {
                    'HTTP_PROXY': f'http://{proxy_url}',
                    'HTTPS_PROXY': f'http://{proxy_url}',
                }
            )

            # display proxy setting
            self.print_setting('Using Proxy Server', f'{self.proxy_host}:{self.proxy_port}')

    def create_requirements_lock(self):
        """Create the requirements.lock file."""
        with Path('requirements.lock').open(mode='w') as fh:
            self.print_setting('Lock File Created', 'requirements.lock')
            fh.write(self.requirements_lock)
            fh.write('')

    def create_temp_requirements(self):
        """Create a temporary requirements.txt.

        This allows testing against a git branch instead of pulling from pypi.
        """
        _requirements_fqfn = Path('requirements.txt')
        if self.has_requirements_lock:
            _requirements_fqfn = Path('requirements.lock')

        # Replace tcex version with develop branch of tcex
        with _requirements_fqfn.open() as fh:
            current_requirements = fh.read().strip().split('\n')

        self.requirements_fqfn_branch = Path(f'temp-{_requirements_fqfn}')
        with self.requirements_fqfn_branch.open(mode='w') as fh:
            requirements = []
            for line in current_requirements:
                if not line:
                    continue
                if line.startswith('tcex'):
                    line = (
                        'git+https://github.com/ThreatConnect-Inc/tcex.git@'
                        f'{self.branch}#egg=tcex'
                    )
                requirements.append(line)
            fh.write('\n'.join(requirements))

        # display branch setting
        self.print_setting('Using Branch', self.branch)

    @property
    def has_requirements_lock(self):
        """Return True if requirements.lock exists."""
        return Path('requirements.lock').exists()

    def install_deps(self):
        """Install Required Libraries using pip."""
        # check for requirements.txt
        if not self.requirements_fqfn.is_file():
            self.handle_error(
                f'A {str(self.requirements_fqfn)} file is required to install modules.'
            )

        # update requirements_dev.txt file
        self.update_requirements_dev_txt()

        # install all requested lib directories
        for lib_version in self.lib_versions:
            # remove lib directory from previous runs
            self._remove_previous(lib_version.lib_dir)

            if (
                not lib_version.python_executable.is_file()
                and not lib_version.python_executable.is_symlink()
            ):

                # display error
                typer.secho(
                    f'The Python executable ({lib_version.python_executable}) could not be found. '
                    'Skipping building lib directory for this Python version.',
                    fg=typer.colors.YELLOW,
                )
                continue

            # display lib dir setting
            self.print_setting('Lib Dir', f'{lib_version.lib_dir.name}')

            # build the sub process command
            exe_command = self._build_command(lib_version.python_executable, lib_version.lib_dir)

            # display command setting
            self.print_setting('Running', f'''{' '.join(exe_command)}''', fg_color='GREEN')

            # recommended -> https://pip.pypa.io/en/latest/user_guide/#using-pip-from-your-program
            p = subprocess.Popen(  # pylint: disable=consider-using-with
                exe_command,
                shell=False,  # nosec
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=self.env,
            )
            _, err = p.communicate()  # pylint: disable=unused-variable

            if p.returncode != 0:
                # display error
                err = err.decode('utf-8')
                failure_display = typer.style(
                    f'Failure: {err}', fg=typer.colors.WHITE, bg=typer.colors.RED
                )
                typer.echo(f'{failure_display}')
                sys.exit(1)

            # TODO: [low] can this be updated to use version from model?
            # version comparison
            try:
                python_version = lib_version.lib_dir.name.split('_', 1)[1]
            except IndexError:
                python_version = None
                self.handle_error('Could not determine version from lib string.')

            # TODO: [low] investigate using sematic_version package
            # track the latest Python version
            if self.latest_version is None or StrictVersion(python_version) > StrictVersion(
                self.latest_version
            ):
                self.latest_version = python_version

        if self.requirements_fqfn_branch:
            # remove temp requirements.txt file
            self.requirements_fqfn_branch.unlink()

        # create lib_latest directory
        self._create_lib_latest()

        if self.has_requirements_lock is False:
            self.create_requirements_lock()

    @property
    def lib_versions(self) -> List[LibVersionModel]:
        """Return the lib_version data required to build lib directories."""
        if self.tj.model.lib_versions:
            self.print_setting('Python Version', 'using version(s) defined in tcex.json')

            # return the python versions defined in the tcex.json file
            return self.tj.model.lib_versions

        # return the current python version
        return [
            LibVersionModel(**{'python_executable': sys.executable, 'lib_dir': self.lib_directory})
        ]

    @cached_property
    def requirements_fqfn(self):
        """Return the appropriate requirements.txt file."""
        if self.dev:
            _requirements_file = Path('requirements_dev.txt')
        elif self.has_requirements_lock:
            _requirements_file = Path('requirements.lock')
        else:
            _requirements_file = Path('requirements.txt')

        self.print_setting('Requirements File', str(_requirements_file))
        return _requirements_file

    @property
    def requirements_lock(self) -> str:
        """Return python packages for requirements.txt."""
        lib_directories = [lib_model.lib_dir for lib_model in self.lib_versions]

        # if we only have one current Python version building the requirements.txt file
        # is straight forward, just include all packages pinned to the current version.
        if len(lib_directories) == 1:
            _requirements = self.requirements_lock_data(lib_directories[0])
            # sort packages alphabetically
            return '\n'.join(sorted(_requirements.splitlines()))

        requirements = {}
        source_requirements = set()
        for lib_dir in lib_directories:
            # extract python major.minor
            python_version = str(lib_dir).replace('lib_', '')
            python_version_major = python_version.split('.', 1)[0]
            python_version_minor = python_version.split('.', 2)[1]

            for requirement in self.requirements_lock_data(lib_dir).splitlines():
                if not requirement:
                    # skip empty lines
                    continue

                try:
                    if ' @ ' in requirement:
                        source_requirements.add(requirement)
                    else:
                        package_name = requirement.split('==')[0]
                        package_version = requirement.split('==')[1]
                        requirements.setdefault(package_name, {})
                        requirements[package_name][
                            f'{python_version_major}.{python_version_minor}'
                        ] = package_version
                except Exception as ex:
                    self.log.error(
                        f'event=get-requirements-txt, requirement={requirement}, error="{ex}"'
                    )
        _requirements = self.requirements_lock_diff(requirements, len(lib_directories))
        _requirements.extend(source_requirements)
        return '\n'.join(sorted(_requirements))

    def requirements_lock_data(self, lib_dir: str) -> str:
        """Return the Python packages for the provided directory."""
        lib_dir_path = os.path.join(self.app_path, lib_dir)
        cmd = f'pip freeze --path "{lib_dir_path}"'
        self.log.debug(f'event=get-requirements-lock-data, cmd={cmd}')
        try:
            output = subprocess.run(  # pylint: disable=subprocess-run-check
                cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE  # nosec
            )
        except Exception as ex:
            self.log.error(f'event=pip-freeze, error="{ex}"')
            typer.echo('Failure: could not get requirements lock data.')
            sys.exit(1)

        if output.returncode != 0:
            self.log.error(f'event=pip-freeze, stderr="{output.stderr}"')

        return output.stdout.decode('utf-8')

    def requirements_lock_diff(self, package_data: dict, python_version_count: int) -> List[str]:
        """Diff the package data, returning the appropriate requirements.lock data."""
        requirements_lock = []
        for package_name, package_version_data in package_data.items():
            if len(package_version_data) < python_version_count:
                # if the package is defined for less than the total number of python versions
                # supported then the package is not required for one or more python versions
                # and must be pinned with the applicable python version defined.
                for python_version, package_version in package_version_data.items():
                    requirements_lock.append(
                        self.requirements_lock_line(package_name, package_version, python_version)
                    )
            elif all(
                x == list(package_version_data.values())[0] for x in package_version_data.values()
            ):
                # if the package version is the same for all python versions then
                # it only needs to be defined once in the requirements.txt file.
                for python_version, package_version in package_version_data.items():
                    requirements_lock.append(
                        self.requirements_lock_line(package_name, package_version)
                    )

                    # only add once
                    break
            else:
                # if the package version is different for one or more versions of python then the
                # python_version must be specified in the requirements.txt file.
                for python_version, package_version in package_version_data.items():
                    requirements_lock.append(
                        self.requirements_lock_line(package_name, package_version, python_version)
                    )
        return requirements_lock

    @staticmethod
    def requirements_lock_line(
        package_name: str, package_version: str, python_version: Optional[str] = None
    ) -> str:
        """Return a string to insert into the requirements.txt file."""
        if python_version is None:
            return f'{package_name}=={package_version}'
        return f'{package_name}=={package_version}; python_version == \'{python_version}\''

    @staticmethod
    def update_requirements_dev_txt():
        """Update the requirements_dev.txt file to support lock file."""
        if os.path.isfile('requirements_dev.txt'):
            with open('requirements_dev.txt', mode='r+') as fh:
                _lines = ''
                for line in fh.readlines():
                    _lines += line.replace('requirements.txt', 'requirements.lock')

                # write back
                fh.seek(0)
                fh.write(_lines)
                fh.truncate()



"""TcEx Utilities Module"""
# standard library
import ast
import ipaddress
import re
from typing import Any, Dict, List, Optional, Pattern, Union

# third-party
import astunparse
import jmespath

# first-party
from tcex.utils.aes_operations import AesOperations
from tcex.utils.datetime_operations import DatetimeOperations
from tcex.utils.string_operations import StringOperations
from tcex.utils.variables import Variables


class Utils(AesOperations, DatetimeOperations, StringOperations, Variables):
    """TcEx Utilities Class"""

    @staticmethod
    def find_line_in_code(
        needle: str,
        code: str,
        trigger_start: Optional[Pattern] = None,
        trigger_stop: Optional[Pattern] = None,
    ) -> str:
        """Return matching line of code in a class definition.

        Args:
            needle: The string to search for.
            code: The contents of the Python file to search.
            trigger_start: The regex pattern to use to trigger the search.
            trigger_stop: The regex pattern to use to stop the search.
        """
        magnet_on = not trigger_start
        for line in astunparse.unparse(ast.parse(code)).split('\n'):
            if line.lstrip()[:1] not in ("'", '"'):
                # Find class before looking for needle
                if trigger_start is not None and re.match(trigger_start, line):
                    magnet_on = True
                    continue

                # find need now that class definition is found
                if magnet_on is True and re.match(needle, line):
                    line = line.strip()
                    return line

                # break if needle not found before next class definition
                if trigger_stop is not None and re.match(trigger_stop, line) and magnet_on is True:
                    break
        return None

    @staticmethod
    def flatten_list(lst: List[Any]) -> List[Any]:
        """Flatten a list

        Will work for lists of lists to arbitrary depth
        and for lists with a mix of lists and single values
        """
        flat_list = []
        for sublist in lst:
            if isinstance(sublist, list):
                for item in Utils.flatten_list(sublist):
                    flat_list.append(item)
            else:
                flat_list.append(sublist)
        return flat_list

    @staticmethod
    def is_cidr(possible_cidr_range: str) -> bool:
        """Return True if the provided value is a valid CIDR block.

        Args:
            possible_cidr_range: The cidr value to validate.
        """
        try:
            ipaddress.ip_address(possible_cidr_range)
        except ValueError:
            try:
                ipaddress.ip_interface(possible_cidr_range)
            except Exception:
                return False
            else:
                return True
        else:
            return False

    @staticmethod
    def is_ip(possible_ip: str) -> bool:
        """Return True if the provided value is a valid IP address.

        Args:
            possible_ip: The IP value to validate.
        """
        try:
            ipaddress.ip_address(possible_ip)
        except ValueError:
            return False
        else:
            return True

    def mapper(self, data: Union[list, dict], mapping: dict):
        """Yield something ..."""
        # TODO [high] - @bpurdy - update docstring with description of what this is?
        if isinstance(data, dict):
            data = [data]
        try:
            for d in data:
                mapped_obj = mapping.copy()
                for key, value in mapping.items():
                    if isinstance(value, list):
                        new_list = []
                        for item in value:
                            if isinstance(item, dict):
                                new_list.append(list(self.mapper(d, item))[0])
                            else:
                                if not item.startswith('@'):
                                    new_list.append(item)
                                else:
                                    new_list.append(
                                        jmespath.search(f'{item}', jmespath.search('@', d))
                                    )

                        mapped_obj[key] = new_list
                    elif isinstance(value, dict):
                        mapped_obj[key] = list(self.mapper(d, mapped_obj[key]))[0]
                    else:
                        if not value.startswith('@'):
                            mapped_obj[key] = value
                        else:
                            mapped_obj[key] = jmespath.search(f'{value}', jmespath.search('@', d))
                yield mapped_obj
        except Exception:  # nosec
            pass

    @staticmethod
    def printable_cred(
        cred: str,
        visible: Optional[int] = 1,
        mask_char: Optional[str] = '*',
        mask_char_count: Optional[int] = 4,
    ) -> str:
        """Return a printable (masked) version of the provided credential.

        Args:
            cred: The cred to print.
            visible: The number of characters at the beginning and ending of the cred to not mask.
            mask_char: The character to use in the mask.
            mask_char_count: How many mask character to insert (obscure cred length).
        """
        if isinstance(cred, str):
            mask_char = mask_char or '*'
            if cred is not None and len(cred) >= visible * 2:
                cred = f'{cred[:visible]}{mask_char * mask_char_count}{cred[-visible:]}'
        return cred

    @staticmethod
    def remove_none(dict_: Dict[Any, Optional[Any]]) -> Dict[Any, Any]:
        """Remove any mappings from a dict with a None value."""
        return {k: v for k, v in dict_.items() if v is not None}

    @staticmethod
    def standardize_asn(asn: str) -> str:
        """Return the ASN formatted for ThreatConnect.

        Args:
            asn: The asn value to standardize.
        """
        numbers = re.findall('[0-9]+', asn)
        if len(numbers) == 1:
            asn = f'ASN{numbers[0]}'
        return asn



"""Gen Code App Input Map"""
# standard library
from typing import List

# first-party
from tcex.app_config.install_json import InstallJson


class SpecToolAppInputStatic:
    """Generate App Config File"""

    def __init__(self):
        """Initialize class properties."""

        # class properties
        self.i1 = ' ' * 4
        self.i2 = ' ' * 8
        self.ij = InstallJson()

    @property
    def template_app_inputs_class(self) -> list:
        """Return app_inputs.py AppInput class."""
        app_model = 'AppBaseModel'
        if self.ij.model.is_trigger_app:
            app_model = 'ServiceConfigModel'

        return [
            '''class AppInputs:''',
            f'''{self.i1}"""App Inputs"""''',
            '',
            f'''{self.i1}def __init__(self, inputs: 'BaseModel'):''',
            f'''{self.i2}"""Initialize class properties."""''',
            f'''{self.i2}self.inputs = inputs''',
            '',
            f'''{self.i1}def update_inputs(self):''',
            f'''{self.i2}"""Add custom App models to inputs.''',
            '',
            f'''{self.i2}Input will be validate when the model is added an any exceptions will''',
            f'''{self.i2}cause the App to exit with a status code of 1.''',
            f'''{self.i2}"""''',
            f'''{self.i2}self.inputs.add_model({app_model})''',
            '',
            '',
        ]

    @property
    def app_base_model_class_comment(self) -> str:
        """Return Service Config Model class."""
        return f'{self.i1}"""Base model for the App containing any common inputs."""'

    @property
    def service_config_model_class_comment(self) -> str:
        """Return Service Config Model class."""
        return '\n'.join(
            [
                f'''{self.i1}"""Base model for the App containing any common inputs.''',
                '',
                f'''{self.i1}Trigger Service App inputs do not take playbookDataType.''',
                '',
                f'''{self.i1}This is the configuration input that is sent to the Service''',
                f'''{self.i1}on startup. The inputs that are configured in the Service''',
                f'''{self.i1}configuration in the Platform with serviceConfig: true''',
                f'''{self.i1}"""''',
                '',
                '',
            ]
        )

    @property
    def trigger_config_model_class_comment(self) -> str:
        """Return Trigger Config Model class."""
        return '\n'.join(
            [
                f'''{self.i1}"""Base model for Trigger (playbook) config.''',
                '',
                f'''{self.i1}Trigger Playbook inputs do not take playbookDataType.''',
                '',
                f'''{self.i1}This is the configuration input that gets sent to the service''',
                f'''{self.i1}when a Playbook is enabled (createConfig).''',
                f'''{self.i1}"""''',
                '',
                '',
            ]
        )

    def template_app_inputs_class_tc_action(self, class_model_map: dict) -> list:
        """Return app_inputs.py AppInput class for App with tc_action."""
        cmm = ''
        for action, class_name in class_model_map.items():
            cmm += f'\'{action}\': {class_name},'
        cmm = f'{{{cmm}}}'

        return [
            '''class AppInputs:''',
            f'''{self.i1}"""App Inputs"""''',
            '',
            f'''{self.i1}def __init__(self, inputs: 'BaseModel'):''',
            f'''{self.i2}"""Initialize class properties."""''',
            f'''{self.i2}self.inputs = inputs''',
            '',
            f'''{self.i1}def get_model(self, tc_action: Optional[str] = None) -> 'BaseModel':''',
            f'''{self.i2}"""Return the model based on the current action."""''',
            f'''{self.i2}tc_action = tc_action or self.inputs.model_unresolved.tc_action''',
            f'''{self.i2}action_model_map = {cmm}''',
            f'''{self.i2}return action_model_map.get(tc_action)''',
            '',
            f'''{self.i1}def update_inputs(self):''',
            f'''{self.i2}"""Add custom App models to inputs.''',
            '',
            f'''{self.i2}Input will be validate when the model is added an any exceptions will''',
            f'''{self.i2}cause the App to exit with a status code of 1.''',
            f'''{self.i2}"""''',
            f'''{self.i2}self.inputs.add_model(self.get_model())''',
            '',
            '',
        ]

    def template_app_inputs_prefix(
        self,
        field_type_modules: List[str],
        pydantic_modules: List[str],
        typing_modules: List[str],
    ) -> list:
        """Return app_inputs.py prefix data."""
        field_types_modules = ', '.join(sorted(list(field_type_modules)))
        pydantic_modules = ', '.join(sorted(list(pydantic_modules)))
        typing_modules = ', '.join(sorted(list(typing_modules)))
        _imports = [
            '"""App Inputs"""',
            '# pylint: disable=no-self-argument, no-self-use',
            '# standard library',
            f'from typing import {typing_modules}',
            '',
            '# third-party',
            f'from pydantic import {pydantic_modules}',
            f'from tcex.input.field_types import {field_types_modules}',
        ]

        # add import for service trigger Apps
        if self.ij.model.is_trigger_app:
            _imports.append('from tcex.input.models import CreateConfigModel')

        # add new lines
        _imports.extend(
            [
                '',
                '',
            ]
        )
        return _imports

    @property
    def type_map(self):
        """Return input map."""
        return {
            'Any': {
                'optional': {'type': 'Any', 'field_type': None},
                'required': {'type': 'Any', 'field_type': None},
            },
            'Binary': {
                'optional': {'type': 'Binary', 'field_type': 'Binary'},
                'required': {'type': 'binary(allow_empty=False)', 'field_type': 'binary'},
            },
            'BinaryArray': {
                'optional': {'type': 'List[Binary]', 'field_type': 'Binary'},
                'required': {'type': 'List[binary(allow_empty=False)]', 'field_type': 'binary'},
            },
            'Choice': {
                'optional': {'type': 'Choice', 'field_type': 'Choice'},
                'required': {'type': 'Choice', 'field_type': 'Choice'},
            },
            'EditChoice': {
                'optional': {'type': 'EditChoice', 'field_type': 'EditChoice'},
                'required': {'type': 'EditChoice', 'field_type': 'EditChoice'},
            },
            'Encrypt': {
                'optional': {'type': 'Sensitive', 'field_type': 'Sensitive'},
                'required': {'type': 'sensitive(allow_empty=False)', 'field_type': 'sensitive'},
            },
            'KeyValue': {
                'optional': {'type': 'KeyValue', 'field_type': 'KeyValue'},
                'required': {'type': 'KeyValue', 'field_type': 'KeyValue'},
            },
            'KeyValueArray': {
                'optional': {'type': 'List[KeyValue]', 'field_type': 'KeyValue'},
                'required': {'type': 'List[KeyValue]', 'field_type': 'KeyValue'},
            },
            'KeyValueList': {
                'optional': {'type': 'List[KeyValue]', 'field_type': 'KeyValue'},
                'required': {'type': 'List[KeyValue]', 'field_type': 'KeyValue'},
            },
            'MultiChoice': {
                'optional': {'type': 'List[Choice]', 'field_type': 'Choice'},
                'required': {'type': 'List[Choice]', 'field_type': 'Choice'},
            },
            'String': {
                'optional': {'type': 'String', 'field_type': 'String'},
                'required': {'type': 'string(allow_empty=False)', 'field_type': 'string'},
            },
            'StringArray': {
                'optional': {'type': 'List[String]', 'field_type': 'String'},
                'required': {'type': 'List[string(allow_empty=False)]', 'field_type': 'string'},
            },
            'TCEntity': {
                'optional': {'type': 'TCEntity', 'field_type': 'TCEntity'},
                'required': {'type': 'TCEntity', 'field_type': 'TCEntity'},
            },
            'TCEntityArray': {
                'optional': {'type': 'List[TCEntity]', 'field_type': 'TCEntity'},
                'required': {'type': 'List[TCEntity]', 'field_type': 'TCEntity'},
            },
        }



"""Install JSON Update"""
# pylint: disable=R0401
# standard library
import os
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:  # pragma: no cover
    from .install_json import InstallJson


class InstallJsonUpdate:
    """Update install.json file with current standards and schema."""

    def __init__(self, ij: 'InstallJson'):  # pylint: disable=E0601
        """Initialize class properties."""
        self.ij = ij

    def multiple(
        self,
        features: Optional[bool] = True,
        migrate: Optional[bool] = False,
        sequence: Optional[bool] = True,
        valid_values: Optional[bool] = True,
        playbook_data_types: Optional[bool] = True,
    ):
        """Update the profile with all required changes.

        Args:
            features: If True, features will be updated.
            migrate: If True, programMain will be set to "run".
            sequence: If True, sequence numbers will be updated.
            valid_values: If True, validValues will be updated.
            playbook_data_types:  If True, pbDataTypes will be updated.
        """
        # update features array
        if features is True:
            self.update_features()

        if migrate is True:
            # update programMain to run
            self.update_program_main()

        # update sequence numbers
        if sequence is True:
            self.update_sequence_numbers()

        # update valid values
        if valid_values is True:
            self.update_valid_values()

        # update playbook data types
        if playbook_data_types is True:
            self.update_playbook_data_types()

        # write updated profile
        self.ij.write()

    # def update_display_name(self, json_data: dict):
    #     """Update the displayName parameter."""
    #     if not self.ij.model.display_name:
    #         display_name = os.path.basename(os.getcwd()).replace(self.app_prefix, '')
    #         display_name = display_name.replace('_', ' ').replace('-', ' ')
    #         display_name = ' '.join([a.title() for a in display_name.split(' ')])
    #     self.ij.model.display_name = self.ij.model.display_name or display_name

    def update_features(self):
        """Update feature set based on App type."""
        features = ['runtimeVariables']

        if self.ij.model.is_organization_app:
            features.extend(['fileParams', 'secureParams'])
        elif self.ij.model.is_playbook_app:
            features.extend(
                [
                    'aotExecutionEnabled',
                    'appBuilderCompliant',
                    'fileParams',
                    'runtimeVariables',
                    'secureParams',
                ]
            )
        elif self.ij.model.is_service_app:
            features.extend(['appBuilderCompliant', 'fileParams'])

        # add layoutEnabledApp if layout.json file exists in project
        if os.path.isfile(os.path.join(self.ij.fqfn.parent, 'layout.json')):  # pragma: no cover
            features.append('layoutEnabledApp')

        # re-add supported optional features
        for feature in self.ij.model.features:
            if feature in [
                'advancedRequest',
                'CALSettings',
                'layoutEnabledApp',
                'smtpSettings',
                'webhookResponseMarshall',
                'webhookServiceEndpoint',
                # features for TC App loop prevention
                'CreatesGroup',
                'CreatesIndicator',
                'CreatesSecurityLabel',
                'CreatesTag',
                'DeletesGroup',
                'DeletesIndicator',
                'DeletesSecurityLabel',
                'DeletesTag',
            ]:
                features.append(feature)

        self.ij.model.features = sorted(list(set(features)))

    def update_program_main(self):
        """Update program main."""
        self.ij.model.program_main = 'run'

    def update_sequence_numbers(self):
        """Update program sequence numbers."""
        for sequence, param in enumerate(self.ij.model.params, start=1):
            param.sequence = sequence

    def update_valid_values(self):
        """Update program main on App type."""
        for param in self.ij.model.params:
            if param.type not in ['String', 'KeyValueList']:
                continue

            store = 'TEXT'
            if param.encrypt is True:
                store = 'KEYCHAIN'

            if self.ij.model.is_organization_app or param.service_config is True:
                if f'${{USER:{store}}}' not in param.valid_values:
                    param.valid_values.append(f'${{USER:{store}}}')

                if f'${{ORGANIZATION:{store}}}' not in param.valid_values:
                    param.valid_values.append(f'${{ORGANIZATION:{store}}}')

                # remove entry that's specifically for playbooks Apps
                if f'${{{store}}}' in param.valid_values:
                    param.valid_values.remove(f'${{{store}}}')

            elif self.ij.model.is_playbook_app:
                if f'${{{store}}}' not in param.valid_values:
                    param.valid_values.append(f'${{{store}}}')

                # remove entry that's specifically for organization (job) Apps
                if f'${{USER:{store}}}' in param.valid_values:
                    param.valid_values.remove(f'${{USER:{store}}}')

                # remove entry that's specifically for organization (job) Apps
                if f'${{ORGANIZATION:{store}}}' in param.valid_values:
                    param.valid_values.remove(f'${{ORGANIZATION:{store}}}')

    def update_playbook_data_types(self):
        """Update program main on App type."""
        if not self.ij.model.is_playbook_app:
            return

        for param in self.ij.model.params:
            if param.type != 'String':
                continue
            if not param.playbook_data_type:
                param.playbook_data_type.append('String')



"""WorkflowEvent / WorkflowEvents Object"""
# standard library
from typing import TYPE_CHECKING, Iterator, Union

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.notes.note_model import NoteModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.workflow_events.workflow_event_filter import WorkflowEventFilter
from tcex.api.tc.v3.workflow_events.workflow_event_model import (
    WorkflowEventModel,
    WorkflowEventsModel,
)

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.notes.note import Note


class WorkflowEvents(ObjectCollectionABC):
    """WorkflowEvents Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = WorkflowEventsModel(**kwargs)
        self.type_ = 'workflow_events'

    def __iter__(self) -> 'WorkflowEvent':
        """Iterate over CM objects."""
        return self.iterate(base_class=WorkflowEvent)

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.WORKFLOW_EVENTS.value

    @property
    def filter(self) -> 'WorkflowEventFilter':
        """Return the type specific filter object."""
        return WorkflowEventFilter(self.tql)


class WorkflowEvent(ObjectABC):
    """WorkflowEvents Object.

    Args:
        case_id (int, kwargs): The **case id** for the Workflow_Event.
        case_xid (str, kwargs): The **case xid** for the Workflow_Event.
        deleted_reason (str, kwargs): The reason for deleting the event (required input for DELETE
            operation only).
        event_date (str, kwargs): The time that the Event is logged.
        notes (Notes, kwargs): A list of Notes corresponding to the Event.
        summary (str, kwargs): The **summary** for the Workflow_Event.
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model = WorkflowEventModel(**kwargs)
        self._nested_field_name = 'workflowEvents'
        self._nested_filter = 'has_workflow_event'
        self.type_ = 'Workflow Event'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.WORKFLOW_EVENTS.value

    @property
    def model(self) -> 'WorkflowEventModel':
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: Union['WorkflowEventModel', dict]):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.summary}

    @property
    def notes(self) -> Iterator['Note']:
        """Yield Note from Notes."""
        # first-party
        from tcex.api.tc.v3.notes.note import Notes

        yield from self._iterate_over_sublist(Notes)

    def stage_note(self, data: Union[dict, 'ObjectABC', 'NoteModel']):
        """Stage note on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = NoteModel(**data)

        if not isinstance(data, NoteModel):
            raise RuntimeError('Invalid type passed in to stage_note')
        data._staged = True
        self.model.notes.data.append(data)



"""Key Value Store module for TcEx Framework"""

# flake8:noqa
from .key_value_api import KeyValueApi
from .key_value_mock import KeyValueMock
from .key_value_redis import KeyValueRedis
from .redis_client import RedisClient



"""App Config Object module for TcEx Framework"""
# flake8: noqa
# first-party
from tcex.app_config.app_spec_yml import AppSpecYml
from tcex.app_config.install_json import InstallJson
from tcex.app_config.job_json import JobJson
from tcex.app_config.layout_json import LayoutJson
from tcex.app_config.tcex_json import TcexJson



"""Note / Notes Model"""
# pylint: disable=no-member,no-self-argument,no-self-use,wrong-import-position
# standard library
from datetime import datetime
from typing import List, Optional

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.utils import Utils


class NotesModel(
    BaseModel,
    title='Notes Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Notes Model"""

    _mode_support = PrivateAttr(False)

    data: Optional[List['NoteModel']] = Field(
        [],
        description='The data for the Notes.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


class NoteDataModel(
    BaseModel,
    title='Note Data Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Notes Data Model"""

    data: Optional[List['NoteModel']] = Field(
        [],
        description='The data for the Notes.',
        methods=['POST', 'PUT'],
        title='data',
    )


class NoteModel(
    V3ModelABC,
    alias_generator=Utils().snake_to_camel,
    extra=Extra.allow,
    title='Note Model',
    validate_assignment=True,
):
    """Note Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(True)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    artifact: Optional['ArtifactModel'] = Field(
        None,
        allow_mutation=False,
        description='The **artifact** for the Note.',
        read_only=True,
        title='artifact',
    )
    artifact_id: Optional[int] = Field(
        None,
        description='The ID of the Artifact on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='artifactId',
    )
    author: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The **author** for the Note.',
        read_only=True,
        title='author',
    )
    case_id: Optional[int] = Field(
        None,
        description='The **case id** for the Note.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseXid',
        title='caseId',
    )
    case_xid: Optional[str] = Field(
        None,
        description='The **case xid** for the Note.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseId',
        title='caseXid',
    )
    date_added: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The **date added** for the Note.',
        read_only=True,
        title='dateAdded',
    )
    edited: bool = Field(
        None,
        allow_mutation=False,
        description='The **edited** for the Note.',
        read_only=True,
        title='edited',
    )
    id: Optional[int] = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    last_modified: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The **last modified** for the Note.',
        read_only=True,
        title='lastModified',
    )
    parent_case: Optional['CaseModel'] = Field(
        None,
        allow_mutation=False,
        description='The **parent case** for the Note.',
        read_only=True,
        title='parentCase',
    )
    summary: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The **summary** for the Note.',
        read_only=True,
        title='summary',
    )
    task: Optional['TaskModel'] = Field(
        None,
        allow_mutation=False,
        description='The **task** for the Note.',
        read_only=True,
        title='task',
    )
    task_id: Optional[int] = Field(
        None,
        description='The ID of the Task on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='taskId',
    )
    task_xid: Optional[str] = Field(
        None,
        description='The XID of the Task on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='taskXid',
    )
    text: Optional[str] = Field(
        None,
        description='The **text** for the Note.',
        methods=['POST', 'PUT'],
        max_length=65500,
        min_length=1,
        read_only=False,
        title='text',
    )
    workflow_event: Optional['WorkflowEventModel'] = Field(
        None,
        allow_mutation=False,
        description='The **workflow event** for the Note.',
        read_only=True,
        title='workflowEvent',
    )
    workflow_event_id: Optional[int] = Field(
        None,
        description='The ID of the Event on which to apply the Note.',
        methods=['POST'],
        read_only=False,
        title='workflowEventId',
    )

    @validator('artifact', always=True)
    def _validate_artifact(cls, v):
        if not v:
            return ArtifactModel()
        return v

    @validator('parent_case', always=True)
    def _validate_case(cls, v):
        if not v:
            return CaseModel()
        return v

    @validator('task', always=True)
    def _validate_task(cls, v):
        if not v:
            return TaskModel()
        return v

    @validator('workflow_event', always=True)
    def _validate_workflow_event(cls, v):
        if not v:
            return WorkflowEventModel()
        return v


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.tasks.task_model import TaskModel
from tcex.api.tc.v3.workflow_events.workflow_event_model import WorkflowEventModel

# add forward references
NoteDataModel.update_forward_refs()
NoteModel.update_forward_refs()
NotesModel.update_forward_refs()



"""Config Models"""

# flake8:noqa
# first-party
from tcex.app_config.models.app_spec_yml_model import AppSpecYmlModel
from tcex.app_config.models.install_json_model import InstallJsonModel
from tcex.app_config.models.job_json_model import JobJsonModel
from tcex.app_config.models.layout_json_model import LayoutJsonModel
from tcex.app_config.models.tcex_json_model import TcexJsonModel
from tcex.app_config.models.template_config_model import TemplateConfigModel



"""User TQL Filter"""
# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class UserFilter(FilterABC):
    """Filter Object for Users"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.USERS.value

    def disabled(self, operator: Enum, disabled: bool):
        """Filter Disabled based on **disabled** keyword.

        Args:
            operator: The operator enum for the filter.
            disabled: A flag indicating whether or not the user's account has been disabled.
        """
        self._tql.add_filter('disabled', operator, disabled, TqlType.BOOLEAN)

    def first_name(self, operator: Enum, first_name: str):
        """Filter First Name based on **firstName** keyword.

        Args:
            operator: The operator enum for the filter.
            first_name: The first name of the user.
        """
        self._tql.add_filter('firstName', operator, first_name, TqlType.STRING)

    def group_id(self, operator: Enum, group_id: int):
        """Filter groupID based on **groupId** keyword.

        Args:
            operator: The operator enum for the filter.
            group_id: The ID of the group the user belongs to.
        """
        self._tql.add_filter('groupId', operator, group_id, TqlType.INTEGER)

    def id(self, operator: Enum, id: int):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the user.
        """
        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def job_function(self, operator: Enum, job_function: str):
        """Filter Job Function based on **jobFunction** keyword.

        Args:
            operator: The operator enum for the filter.
            job_function: The user's job function.
        """
        self._tql.add_filter('jobFunction', operator, job_function, TqlType.STRING)

    def job_role(self, operator: Enum, job_role: str):
        """Filter Job Role based on **jobRole** keyword.

        Args:
            operator: The operator enum for the filter.
            job_role: The user's job role.
        """
        self._tql.add_filter('jobRole', operator, job_role, TqlType.STRING)

    def last_login(self, operator: Enum, last_login: str):
        """Filter Last Login based on **lastLogin** keyword.

        Args:
            operator: The operator enum for the filter.
            last_login: The last time the user logged in.
        """
        last_login = self.utils.any_to_datetime(last_login).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastLogin', operator, last_login, TqlType.STRING)

    def last_name(self, operator: Enum, last_name: str):
        """Filter Last Name based on **lastName** keyword.

        Args:
            operator: The operator enum for the filter.
            last_name: The last name of the user.
        """
        self._tql.add_filter('lastName', operator, last_name, TqlType.STRING)

    def last_password_change(self, operator: Enum, last_password_change: str):
        """Filter Last Password Change based on **lastPasswordChange** keyword.

        Args:
            operator: The operator enum for the filter.
            last_password_change: The last time the user changed their password.
        """
        last_password_change = self.utils.any_to_datetime(last_password_change).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('lastPasswordChange', operator, last_password_change, TqlType.STRING)

    def locked(self, operator: Enum, locked: bool):
        """Filter Locked based on **locked** keyword.

        Args:
            operator: The operator enum for the filter.
            locked: A flag indicating whether or not the user's account has been locked.
        """
        self._tql.add_filter('locked', operator, locked, TqlType.BOOLEAN)

    def logout_interval_minutes(self, operator: Enum, logout_interval_minutes: int):
        """Filter Logout Interval based on **logoutIntervalMinutes** keyword.

        Args:
            operator: The operator enum for the filter.
            logout_interval_minutes: The configured period of time to wait for idle activity before
                being logged out.
        """
        self._tql.add_filter(
            'logoutIntervalMinutes', operator, logout_interval_minutes, TqlType.INTEGER
        )

    def password_reset_required(self, operator: Enum, password_reset_required: bool):
        """Filter Password Reset Required based on **passwordResetRequired** keyword.

        Args:
            operator: The operator enum for the filter.
            password_reset_required: A flag indicating whether or not the user's password needs to
                be reset upoin next login.
        """
        self._tql.add_filter(
            'passwordResetRequired', operator, password_reset_required, TqlType.BOOLEAN
        )

    def pseudonym(self, operator: Enum, pseudonym: str):
        """Filter Pseudonym based on **pseudonym** keyword.

        Args:
            operator: The operator enum for the filter.
            pseudonym: The user's pseudonym.
        """
        self._tql.add_filter('pseudonym', operator, pseudonym, TqlType.STRING)

    def system_role(self, operator: Enum, system_role: str):
        """Filter Role Name based on **systemRole** keyword.

        Args:
            operator: The operator enum for the filter.
            system_role: The system role name defined for the user.
        """
        self._tql.add_filter('systemRole', operator, system_role, TqlType.STRING)

    def terms_accepted(self, operator: Enum, terms_accepted: bool):
        """Filter Terms Accepted based on **termsAccepted** keyword.

        Args:
            operator: The operator enum for the filter.
            terms_accepted: Flag indicating whether user has accepted the current Terms of Service.
        """
        self._tql.add_filter('termsAccepted', operator, terms_accepted, TqlType.BOOLEAN)

    def terms_accepted_date(self, operator: Enum, terms_accepted_date: str):
        """Filter Terms Accepted Date based on **termsAcceptedDate** keyword.

        Args:
            operator: The operator enum for the filter.
            terms_accepted_date: Date and time the user accepted the current Terms of Service.
        """
        terms_accepted_date = self.utils.any_to_datetime(terms_accepted_date).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('termsAcceptedDate', operator, terms_accepted_date, TqlType.STRING)

    def tql_timeout(self, operator: Enum, tql_timeout: int):
        """Filter TQL Timeout based on **tqlTimeout** keyword.

        Args:
            operator: The operator enum for the filter.
            tql_timeout: The custom TQL timeout value (if defined).
        """
        self._tql.add_filter('tqlTimeout', operator, tql_timeout, TqlType.INTEGER)

    def ui_theme(self, operator: Enum, ui_theme: str):
        """Filter UI Theme based on **uiTheme** keyword.

        Args:
            operator: The operator enum for the filter.
            ui_theme: The user's configured theme (e.g. light/dark).
        """
        self._tql.add_filter('uiTheme', operator, ui_theme, TqlType.STRING)

    def user_name(self, operator: Enum, user_name: str):
        """Filter User Name based on **userName** keyword.

        Args:
            operator: The operator enum for the filter.
            user_name: The user name of the user.
        """
        self._tql.add_filter('userName', operator, user_name, TqlType.STRING)



"""Input Models"""

# flake8:noqa
# first-party
from tcex.input.models.advanced_request_model import _AdvancedRequestModel
from tcex.input.models.aot_execution_enabled_model import AotExecutionEnabledModel
from tcex.input.models.api_model import ApiModel
from tcex.input.models.batch_model import BatchModel
from tcex.input.models.cal_settings_model import CalSettingsModel
from tcex.input.models.create_config_model import CreateConfigModel
from tcex.input.models.logging_model import LoggingModel
from tcex.input.models.model_map import feature_map, runtime_level_map, tc_action_map
from tcex.input.models.organization_model import OrganizationModel
from tcex.input.models.path_model import PathModel
from tcex.input.models.playbook_common_model import PlaybookCommonModel
from tcex.input.models.playbook_model import PlaybookModel
from tcex.input.models.proxy_model import ProxyModel
from tcex.input.models.service_model import ServiceModel
from tcex.input.models.smtp_settings_model import SmtpSettingsModel
